from __future__ import print_function

import numpy as np
import math
import pandas as pd
import scipy.stats as ss
import scipy.stats as stats
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

from sklearn.mixture import GaussianMixture
from scipy.signal import savgol_filter
from scipy.signal import find_peaks
from piscat.Trajectory.data_handeling import protein_trajectories_list2dic
from piscat.InputOutput import read_write_data
from piscat.Visualization.print_colors import PrintColors
from piscat.Visualization.plot import plot2df, plot_bright_dark_psf, plot_bright_dark_psf_inTime
from tqdm.autonotebook import tqdm


class PlotProteinHistogram(PrintColors):

    def __init__(self, intersection_display_flag=False, imgSizex=5, imgSizey=5, flag_localization_filter=False, radius=None):

        """
        This class use video analysis data ('HDF5', 'Matlab') to plot histograms.

        Parameters
        ----------
        intersection_display_flag: bool
            This flag can be used when we want to see the result of intersection on
            top of v-shaped (Please read tutorials 3).

        imgSizex: int
            The width of the histogram figure.

        imgSizey: int
            The height of the histogram figure.

        flag_localization_filter: bool
            This flag is used to define a mask and filter PSF depending on the localization map.

        radius: int
            This parameter is used to define the radius of a circular mask that has the same center as the localization
            map and filters PSF in the edges and border.

        """
        PrintColors.__init__(self)

        self.radius = radius
        self.flag_localization_filter = flag_localization_filter
        self.centerOfImage_X = None
        self.centerOfImage_Y = None

        self.imgSizex = imgSizex
        self.imgSizey = imgSizey

        self.intersection_display_flag = intersection_display_flag
        self.folder_name = []

        self.t_particle_center_intensity = []
        self.t_particle_center_intensity_follow = []

        self.tSmooth_particle_center_intensity = []
        self.tSmooth_particle_center_intensity_follow = []

        self.t_contrast_peaks = []
        self.t_contrast_peaks_index = []

        self.t_contrast_intersection = []
        self.t_contrast_intersection_index = []

        self.t_contrast_Prominence = []
        self.t_contrast_Prominence_index = []

        self.t_iPSFCentroidSigmas_ = []

        self.t_len_linking = []

        self.t_particle_center_fit_intensity = []

        self.tSmooth_particle_center_intensity = []

        self.t_contrast_fit_peaks = []
        self.t_contrast_fit_peaks_index = []

        self.t_contrast_fit_intersection = []
        self.t_contrast_fit_intersection_index = []

        self.t_iPSFCentroidSigmas_fit_x = []
        self.t_iPSFCentroidSigmas_fit_y = []

        self.t_mean_x_center_bright = []
        self.t_std_x_center_bright = []
        self.t_mean_y_center_bright = []
        self.t_std_y_center_bright = []
        self.t_particle_ID_bright = []
        self.t_particle_frame_bright = []
        self.t_linking_len_bright = []

        self.t_mean_x_center_dark = []
        self.t_std_x_center_dark = []
        self.t_mean_y_center_dark = []
        self.t_std_y_center_dark = []
        self.t_particle_ID_dark = []
        self.t_particle_frame_dark = []
        self.t_linking_len_dark = []

        self.t_particle_localization_flag = []

    def __call__(self, folder_name, particles, batch_size, video_frame_num, MinPeakWidth, MinPeakProminence,
                 pixel_size=0.66):
        """
        By calling the object of class this function tries to read the result of the corresponding video it is defined
        with ``folder_name``. These results concatenated with previous results to use for plotting histogram.

        Parameters
        ----------
        folder_name: str
            Name of a folder that data read from it. This saves next to data for convinces tracking the analysis.

        particles: list/dict

            | [intensity_horizontal, intensity_vertical, particle_center_intensity,
                            particle_center_intensity_follow, particle_frame, particle_sigma, particle_X, particle_Y, particle_ID,
                            optional(fit_intensity, fit_x, fit_y, fit_X_sigma, fit_Y_sigma, fit_Bias, fit_intensity_error,
                            fit_x_error, fit_y_error, fit_X_sigma_error, fit_Y_sigma_error, fit_Bias_error)]

            | {"#0": {'intensity_horizontal': ..., 'intensity_vertical': ..., ..., 'particle_ID': ...},
                        "#1": {}, ...}

        batch_size: int
            The size of batch that was used on the DRA step.

        video_frame_num: int
            The number of frames for the corresponding video.

        MinPeakWidth: int
            This is defined as the minimum V-shaped mouth that will use for prominence.

        MinPeakProminence: int
            This is defined as the minimum V-shape height that will use for prominence.

        pixel_size: float
            The size of the camera pixel.

        """

        if type(particles) is list:
            particles = np.asarray(particles)

        if type(particles) is dict:
            for key, item in particles.items():
                intensity_horizontal = item['intensity_horizontal']
                intensity_vertical = item['intensity_vertical']
                center_int = item['center_int']
                center_int_flow = item['center_int_flow']
                frame_number = item['frame_number']
                sigma = item['sigma']
                x_center = item['x_center']
                y_center = item['y_center']
                particle_ID = item['particle_ID']

                num_parameters = len(item)

                if num_parameters == 21:
                    fit_intensity = item['fit_intensity']
                    fit_x = item['fit_x'] * pixel_size
                    fit_y = item['fit_y'] * pixel_size
                    fit_X_sigma = item['fit_X_sigma'] * pixel_size
                    fit_Y_sigma = item['fit_Y_sigma'] * pixel_size
                    fit_Bias = item['fit_Bias']
                    fit_intensity_error = item['fit_intensity_error']
                    fit_x_error = item['fit_x_error'] * pixel_size
                    fit_y_error = item['fit_y_error'] * pixel_size
                    fit_X_sigma_error = item['fit_X_sigma_error'] * pixel_size
                    fit_Y_sigma_error = item['fit_Y_sigma_error'] * pixel_size
                    fit_Bias_error = item['fit_Bias_error']

                else:
                    fit_intensity = None
                    fit_X_sigma = None
                    fit_Y_sigma = None

                # self.localization(x_center, y_center, center_int, key, frame_number)
                center_int_clean = center_int[~np.isnan(center_int)]
                particle_ID_clean = particle_ID[~np.isnan(center_int)]
                frame_number_clean = frame_number[~np.isnan(center_int)]
                x_center_clean = x_center[~np.isnan(center_int)]
                y_center_clean = y_center[~np.isnan(center_int)]
                center_int_flow_clean = center_int_flow[~np.isnan(center_int_flow)]
                sigma_clean = sigma[~np.isnan(center_int)]

                self.localization(x_center_clean, y_center_clean, center_int_clean, particle_ID_clean, frame_number_clean)

                self.data_handling(center_int_clean, center_int_flow_clean, folder_name, sigma_clean, num_parameters, fit_intensity,
                                   fit_X_sigma, fit_Y_sigma, frame_number, batch_size, video_frame_num,
                                   MinPeakWidth=MinPeakWidth, MinPeakProminence=MinPeakProminence)

        if type(particles) is np.ndarray:

            for n_particle in range(particles.shape[0]):
                if type(particles[n_particle][0]) is list:
                    intensity_horizontal = np.asarray(particles[n_particle][0])
                    intensity_vertical = np.asarray(particles[n_particle][1])
                    center_int = np.asarray(particles[n_particle][2])
                    center_int_flow = np.asarray(particles[n_particle][3])
                    frame_number = np.asarray(particles[n_particle][4])
                    sigma = np.asarray(particles[n_particle][5])
                    x_center = np.asarray(particles[n_particle][6])
                    y_center = np.asarray(particles[n_particle][7])
                    particle_ID = np.asarray(particles[n_particle][8])
                else:
                    intensity_horizontal = particles[n_particle][0].ravel()
                    intensity_vertical = particles[n_particle][1].ravel()
                    center_int = particles[n_particle][2].ravel()
                    center_int_flow = particles[n_particle][3].ravel()
                    frame_number = particles[n_particle][4].ravel()
                    sigma = particles[n_particle][5].ravel()
                    x_center = particles[n_particle][6].ravel()
                    y_center = particles[n_particle][7].ravel()
                    particle_ID = particles[n_particle][8].ravel()

                num_parameters = len(particles[n_particle])

                if num_parameters == 21:
                    if type(particles[n_particle][0]) is list:
                        fit_intensity = np.asarray(particles[n_particle][9])
                        fit_x = np.asarray(particles[n_particle][10]) * pixel_size
                        fit_y = np.asarray(particles[n_particle][11]) * pixel_size
                        fit_X_sigma = np.asarray(particles[n_particle][12]) * pixel_size
                        fit_Y_sigma = np.asarray(particles[n_particle][13]) * pixel_size
                        fit_Bias = np.asarray(particles[n_particle][14])
                        fit_intensity_error = np.asarray(particles[n_particle][15])
                        fit_x_error = np.asarray(particles[n_particle][16]) * pixel_size
                        fit_y_error = np.asarray(particles[n_particle][17]) * pixel_size
                        fit_X_sigma_error = np.asarray(particles[n_particle][18]) * pixel_size
                        fit_Y_sigma_error = np.asarray(particles[n_particle][19]) * pixel_size
                        fit_Bias_error = np.asarray(particles[n_particle][20])

                    else:
                        fit_intensity = particles[n_particle][9].ravel()
                        fit_x = particles[n_particle][10].ravel() * pixel_size
                        fit_y = particles[n_particle][11].ravel() * pixel_size
                        fit_X_sigma = particles[n_particle][12].ravel() * pixel_size
                        fit_Y_sigma = particles[n_particle][13].ravel() * pixel_size
                        fit_Bias = particles[n_particle][14].ravel()
                        fit_intensity_error = particles[n_particle][15].ravel()
                        fit_x_error = particles[n_particle][16].ravel() * pixel_size
                        fit_y_error = particles[n_particle][17].ravel() * pixel_size
                        fit_X_sigma_error = particles[n_particle][18].ravel() * pixel_size
                        fit_Y_sigma_error = particles[n_particle][19].ravel() * pixel_size
                        fit_Bias_error = particles[n_particle][20].ravel()

                else:
                    fit_intensity = None
                    fit_X_sigma = None
                    fit_Y_sigma = None

                # self.localization(x_center, y_center, center_int, key, frame_number)
                center_int_clean = center_int[~np.isnan(center_int)]
                particle_ID_clean = particle_ID[~np.isnan(center_int)]
                frame_number_clean = frame_number[~np.isnan(center_int)]
                x_center_clean = x_center[~np.isnan(center_int)]
                y_center_clean = y_center[~np.isnan(center_int)]
                center_int_flow_clean = center_int_flow[~np.isnan(center_int_flow)]
                sigma_clean = sigma[~np.isnan(center_int)]

                self.localization(x_center_clean, y_center_clean, center_int_clean, particle_ID_clean,
                                  frame_number_clean)

                self.data_handling(center_int_clean, center_int_flow_clean, folder_name, sigma_clean, num_parameters,
                                   fit_intensity,
                                   fit_X_sigma, fit_Y_sigma, frame_number, batch_size, video_frame_num,
                                   MinPeakWidth=MinPeakWidth, MinPeakProminence=MinPeakProminence)

    def get_setting(self, flag_localization_filter, radius, centerOfImage_X, centerOfImage_Y):
        self.radius = radius
        self.centerOfImage_X = centerOfImage_X
        self.centerOfImage_Y = centerOfImage_Y
        self.flag_localization_filter = flag_localization_filter

    def localization(self, x_center, y_center, center_int, particle_ID, frame_number):
        if np.mean(center_int) >= 0:
            if self.flag_localization_filter:
                x = np.mean(x_center) - self.centerOfImage_X
                y = np.mean(y_center) - self.centerOfImage_Y
                if np.sqrt(x**2 + y**2) <= self.radius:

                        self.t_particle_localization_flag = True

                        self.t_mean_x_center_bright.append(np.mean(x_center))
                        self.t_std_x_center_bright.append(np.std(x_center))
                        self.t_mean_y_center_bright.append(np.mean(y_center))
                        self.t_std_y_center_bright.append(np.std(y_center))
                        self.t_particle_ID_bright.append(np.unique(particle_ID)[0])
                        self.t_particle_frame_bright.append(int(np.median(frame_number)))
                        self.t_linking_len_bright.append(len(center_int))

                else:
                    self.t_particle_localization_flag = False
            else:
                self.t_particle_localization_flag = True

                self.t_mean_x_center_bright.append(np.mean(x_center))
                self.t_std_x_center_bright.append(np.std(x_center))
                self.t_mean_y_center_bright.append(np.mean(y_center))
                self.t_std_y_center_bright.append(np.std(y_center))
                self.t_particle_ID_bright.append(np.unique(particle_ID)[0])
                self.t_particle_frame_bright.append(int(np.median(frame_number)))
                self.t_linking_len_bright.append(len(center_int))

        elif np.mean(center_int) < 0:

            if self.flag_localization_filter:
                x = np.mean(x_center) - self.centerOfImage_X
                y = np.mean(y_center) - self.centerOfImage_Y
                if np.sqrt(x**2 + y**2) <= self.radius:

                        self.t_particle_localization_flag = True

                        self.t_mean_x_center_dark.append(np.mean(x_center))
                        self.t_std_x_center_dark.append(np.std(x_center))
                        self.t_mean_y_center_dark.append(np.mean(y_center))
                        self.t_std_y_center_dark.append(np.std(y_center))
                        self.t_particle_ID_dark.append(np.unique(particle_ID)[0])
                        self.t_particle_frame_dark.append(int(np.median(frame_number)))
                        self.t_linking_len_dark.append(len(center_int))

                else:
                    self.t_particle_localization_flag = False
            else:
                self.t_particle_localization_flag = True

                self.t_mean_x_center_dark.append(np.mean(x_center))
                self.t_std_x_center_dark.append(np.std(x_center))
                self.t_mean_y_center_dark.append(np.mean(y_center))
                self.t_std_y_center_dark.append(np.std(y_center))
                self.t_particle_ID_dark.append(np.unique(particle_ID)[0])
                self.t_particle_frame_dark.append(int(np.median(frame_number)))
                self.t_linking_len_dark.append(len(center_int))

    def data_handling(self, center_int, center_int_flow, folder_name, sigma, num_parameters, fit_intensity,
                      fit_X_sigma, fit_Y_sigma, frame_number, batch_size, video_frame_num, MinPeakWidth, MinPeakProminence):

        if len(center_int) != 0 and ~np.isnan(center_int).all() and self.t_particle_localization_flag:
            win_size = self.determine_windows_size(center_int)
            if win_size > 3:
                V_smooth = savgol_filter(center_int, win_size, 3)

                win_size = self.determine_windows_size(center_int_flow)
                V_smooth_follow = savgol_filter(center_int_flow, win_size, 3)

                peak, index_peak, idx_ = self.find_peack_contrast(frame_number, V_smooth)
                xi, yi = self.intersection_point(data_index=frame_number, data=V_smooth, index=index_peak,
                                                 display=self.intersection_display_flag)

                tprofile_frameNo_DRA_longer, start_FN_earlyV, end_FN_earlyV = self.frameNumber_longerProfile_2_Vshape(batch_size,
                                                                                                                      video_frame_num,
                                                                                                                      frame_number,
                                                                                                                      V_smooth_follow)

                prom, prom_idx, p_idx_, pro_ = self.find_Prominence_contrast(tprofile_DRA_smooth=V_smooth, tprofile_DRA_tail_smooth=V_smooth_follow,
                                                               tprofile_frameNo_DRA_longer=tprofile_frameNo_DRA_longer,
                                                               start_FN_earlyV=start_FN_earlyV, end_FN_earlyV=end_FN_earlyV,
                                                               MinPeakWidth=MinPeakWidth, MinPeakProminence=MinPeakProminence)

                properties = [V_smooth, V_smooth_follow, [peak, index_peak], [yi, xi], [prom, prom_idx, p_idx_, pro_],
                              tprofile_frameNo_DRA_longer, start_FN_earlyV, end_FN_earlyV]

                self.folder_name.append(folder_name)

                self.t_particle_center_intensity.append(center_int)
                self.t_particle_center_intensity_follow.append(center_int_flow)

                self.tSmooth_particle_center_intensity.append(V_smooth)
                self.tSmooth_particle_center_intensity_follow.append(V_smooth_follow)

                self.t_contrast_peaks.append(peak)
                self.t_contrast_peaks_index.append(index_peak)

                self.t_contrast_intersection.append(yi)
                self.t_contrast_intersection_index.append(xi)
                self.t_len_linking.append(len(center_int))

                self.t_contrast_Prominence.append(prom)
                self.t_contrast_Prominence_index.append(prom_idx)

                self.t_iPSFCentroidSigmas_.append(sigma[idx_])
            else:
                properties = None
        else:
            properties = None

        if num_parameters == 21 and self.t_particle_localization_flag:
            nan_array = np.isnan(fit_intensity)
            not_nan_array = ~ nan_array
            fit_intensity_ = fit_intensity[not_nan_array]

            win_size = self.determine_windows_size(fit_intensity_)
            if len(fit_intensity_) != 0:
                if win_size > 3:
                    V_smooth = savgol_filter(fit_intensity_, win_size, 3)

                    peak, index_peak, idx_ = self.find_peack_contrast(frame_number, V_smooth)
                    xi, yi = self.intersection_point(data_index=frame_number, data=V_smooth, index=index_peak,
                                                     display=False)

                    self.t_particle_center_fit_intensity.append(fit_intensity_)
                    self.tSmooth_particle_center_intensity.append(V_smooth)

                    self.t_contrast_fit_peaks.append(peak)
                    self.t_contrast_fit_peaks_index.append(index_peak)

                    self.t_contrast_fit_intersection.append(yi)
                    self.t_contrast_fit_intersection_index.append(xi)

                    fit_X_sigma_ = fit_X_sigma[not_nan_array]
                    fit_Y_sigma_ = fit_Y_sigma[not_nan_array]

                    self.t_iPSFCentroidSigmas_fit_x.append(fit_X_sigma_[idx_])
                    self.t_iPSFCentroidSigmas_fit_y.append(fit_Y_sigma_[idx_])
                    fit_properties = [V_smooth, V_smooth_follow, [peak, index_peak], [yi, xi], [prom, prom_idx]]

                    return properties, fit_properties

                else:
                    return properties, None

            else:
                return properties, None

        else:
            return properties, None

    def determine_windows_size(self, signal):

        window_size = round(0.05 * len(signal))
        if (window_size % 2) == 0:
            window_size = window_size + 1
        return window_size

    def find_peack_contrast(self, tprofile_DRA_index, tprofile_DRA):
        if np.mean(tprofile_DRA) >= 0:
            max_center_intensity = np.max(tprofile_DRA)
            index_max_ = np.argmax(tprofile_DRA)
            index_max_center_intensity = tprofile_DRA_index[index_max_]
            return max_center_intensity, index_max_center_intensity, index_max_

        elif np.mean(tprofile_DRA) < 0:
            min_center_intensity = np.min(tprofile_DRA)
            index_min_ = np.argmin(tprofile_DRA)
            index_min_center_intensity = tprofile_DRA_index[index_min_]
            return min_center_intensity, index_min_center_intensity, index_min_

    def find_Prominence_contrast(self, tprofile_DRA_smooth, tprofile_DRA_tail_smooth, tprofile_frameNo_DRA_longer,
                                 start_FN_earlyV, end_FN_earlyV, MinPeakWidth, MinPeakProminence):

        if np.mean(tprofile_DRA_smooth) >= 0:
            peaks_index, properties = find_peaks(x=tprofile_DRA_tail_smooth, width=[MinPeakWidth, np.inf],
                                                 prominence=[MinPeakProminence, np.inf])
            peaks_index = peaks_index.tolist()
            frame_index_ = [tprofile_frameNo_DRA_longer[peaks_index[i_]] for i_ in range(len(peaks_index))]

            idx_ = []
            frame_idx_ = []
            pro_properties = []
            for i_ in range(len(frame_index_)):
                f_idx = frame_index_[i_]
                if f_idx <= end_FN_earlyV and f_idx >= start_FN_earlyV:
                    idx_.append(peaks_index[i_])
                    frame_idx_.append(f_idx)
                    pro_properties.append(properties["prominences"][i_])

            if len(idx_) != 0:
                prom_ = [tprofile_DRA_tail_smooth[i_] for i_ in idx_]
                prom = np.min(prom_)
                prom_idx = np.argmin(prom_)
                prom_index = frame_idx_[prom_idx]
                pro = pro_properties[prom_idx]

            else:
                prom = np.nan
                prom_index = np.nan
                prom_idx = np.nan
                pro = np.nan

        elif np.mean(tprofile_DRA_smooth) < 0:
            peaks_index, properties = find_peaks(x=-1*tprofile_DRA_tail_smooth, width=[MinPeakWidth, np.inf],
                                                 prominence=[MinPeakProminence, np.inf])
            peaks_index = peaks_index.tolist()
            frame_index_ = [tprofile_frameNo_DRA_longer[peaks_index[i_]] for i_ in range(len(peaks_index))]

            idx_ = []
            frame_idx_ = []
            pro_properties = []
            for i_ in range(len(frame_index_)):
                f_idx = frame_index_[i_]
                if f_idx <= end_FN_earlyV and f_idx >= start_FN_earlyV:
                    idx_.append(peaks_index[i_])
                    frame_idx_.append(f_idx)
                    pro_properties.append(properties["prominences"][i_])

            if len(idx_) != 0:
                prom_ = [tprofile_DRA_tail_smooth[i_] for i_ in idx_]
                prom = -1 * np.min(np.abs(prom_))
                prom_idx = np.argmin(prom_)
                prom_index = frame_idx_[prom_idx]
                pro = pro_properties[prom_idx]
            else:
                prom = np.nan
                prom_index = np.nan
                prom_idx = np.nan
                pro = np.nan

        return prom, prom_index, prom_idx, pro

    def intersection_point(self, data_index, data, index, display=False):
        try:
            data = data.ravel()
            idx_ = np.where(data_index == index)
            right_side = data[0:idx_[0][0]]
            left_side = data[idx_[0][0]:]

            data_index = np.asarray(data_index)
            x_right = data_index[data_index < index]
            x_left = data_index[data_index >= index]

            z_right = np.polyfit(x_right, right_side, 1, full=True)
            p_l_right = np.poly1d(z_right[0])
            y_right = p_l_right[0] + p_l_right[1] * x_right

            z_left = np.polyfit(x_left, left_side, 1, full=True)
            p_l_left = np.poly1d(z_left[0])
            y_left = p_l_left[0] + p_l_left[1] * x_left

            if p_l_right[1] == p_l_left[1]:
                raise Exception('---lines do not intersect!---')
            else:
                xi = (p_l_left[0] - p_l_right[0]) / (p_l_right[1] - p_l_left[1])
                yi = p_l_left[1] * xi + p_l_left[0]

            if display:
                self.fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
                self.axs = self.fig.add_axes([0, 0, 1, 1])
                self.axs.plot(x_right, y_right, 'r', label='Fitted line at right', zorder=1)
                self.axs.plot(x_left, y_left, 'r', label='Fitted line at left', zorder=2)
                self.axs.scatter(xi, yi, s=80, facecolors='none', edgecolors='k', zorder=3, label='Intersection of the lines')
                self.axs.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))

            return int(xi), yi

        except:
            xi = np.nan
            yi = np.nan

            return xi, yi

    def mean_std(self, input):
        input = np.abs(input)
        if input.shape[0] != 0:
            std_ = np.nanstd(input)
            mean_ = np.nanmean(input)
        else:
            std_ = 0
            mean_ = 0
        return mean_, std_

    def extract_hist_information(self, con_intersections, con_peaks, con_proms, upper_limitation=1, lower_limitation=-1, max_n_components=3, Flag_GMM_fit=True):
        t_contrast_intersection = np.asarray(con_intersections)
        t_contrast_intersection = t_contrast_intersection[~np.isnan(t_contrast_intersection)]

        t_bright_contrast_intersection = t_contrast_intersection[t_contrast_intersection >= 0]
        t_black_contrast_intersection = t_contrast_intersection[t_contrast_intersection < 0]

        t_contrast_peaks = np.asarray(con_peaks)
        t_contrast_peaks = t_contrast_peaks[~np.isnan(t_contrast_peaks)]

        t_bright_contrast_peaks = t_contrast_peaks[t_contrast_peaks >= 0]
        t_black_contrast_peaks = t_contrast_peaks[t_contrast_peaks < 0]
        if con_proms is not None:
            t_contrast_proms = np.asarray(con_proms)
            t_contrast_proms = t_contrast_proms[~np.isnan(t_contrast_proms)]

            t_bright_contrast_proms = t_contrast_proms[t_contrast_proms >= 0]
            t_black_contrast_proms = t_contrast_proms[t_contrast_proms < 0]

        # lower limit
        t_bright_contrast_intersection = t_bright_contrast_intersection[t_bright_contrast_intersection >= lower_limitation]
        t_black_contrast_intersection = t_black_contrast_intersection[t_black_contrast_intersection >= lower_limitation]

        t_bright_contrast_peaks = t_bright_contrast_peaks[t_bright_contrast_peaks >= lower_limitation]
        t_black_contrast_peaks = t_black_contrast_peaks[t_black_contrast_peaks >= lower_limitation]

        if con_proms is not None:
            t_bright_contrast_proms = t_bright_contrast_proms[t_bright_contrast_proms >= lower_limitation]
            t_black_contrast_proms = t_black_contrast_proms[t_black_contrast_proms >= lower_limitation]
            t_contrast_proms = t_contrast_proms[t_contrast_proms >= lower_limitation]

        t_contrast_intersection = t_contrast_intersection[t_contrast_intersection >= lower_limitation]
        t_contrast_peaks = t_contrast_peaks[t_contrast_peaks >= lower_limitation]

        # upper limit
        t_bright_contrast_intersection = t_bright_contrast_intersection[t_bright_contrast_intersection < upper_limitation]
        t_black_contrast_intersection = t_black_contrast_intersection[t_black_contrast_intersection < upper_limitation]

        t_bright_contrast_peaks = t_bright_contrast_peaks[t_bright_contrast_peaks < upper_limitation]
        t_black_contrast_peaks = t_black_contrast_peaks[t_black_contrast_peaks < upper_limitation]

        if con_proms is not None:
            t_bright_contrast_proms = t_bright_contrast_proms[t_bright_contrast_proms < upper_limitation]
            t_black_contrast_proms = t_black_contrast_proms[t_black_contrast_proms < upper_limitation]
            t_contrast_proms = t_contrast_proms[t_contrast_proms < upper_limitation]

        t_contrast_intersection = t_contrast_intersection[t_contrast_intersection < upper_limitation]
        t_contrast_peaks = t_contrast_peaks[t_contrast_peaks < upper_limitation]

        # std & mean
        mean_bright_intersection, std_bright_intersection = self.mean_std(t_bright_contrast_intersection)
        mean_black_intersection, std_black_intersection = self.mean_std(t_black_contrast_intersection)

        mean_bright_peak, std_bright_peak = self.mean_std(t_bright_contrast_peaks)
        mean_black_peak, std_black_peak = self.mean_std(t_black_contrast_peaks)

        if con_proms is not None:
            mean_bright_prom, std_bright_prom = self.mean_std(t_bright_contrast_proms)
            mean_black_prom, std_black_prom = self.mean_std(t_black_contrast_proms)
            mean_prom, std_prom = self.mean_std(t_contrast_proms)

        mean_intersection, std_intersection = self.mean_std(t_contrast_intersection)
        mean_peak, std_peak = self.mean_std(t_contrast_peaks)

        sci_num = lambda x: "{:.2e}".format(x)

        if con_proms is not None:

            list_data = [t_bright_contrast_peaks, t_bright_contrast_intersection, t_bright_contrast_proms,
                         t_black_contrast_peaks, t_black_contrast_intersection, t_black_contrast_proms,
                         t_contrast_peaks, t_contrast_intersection, t_contrast_proms]

            title = ['Peak bright', 'Intersection bright', 'Prominence bright',
                     'Peak black', 'Intersection black', 'Prominence black',
                     'Total peak', 'Total Intersection', 'Total Prominence']

            dic = {'Peak bright': [sci_num(mean_bright_peak), sci_num(std_bright_peak), int(t_bright_contrast_peaks.shape[0])],
                   'Peak black': [sci_num(mean_black_peak), sci_num(std_black_peak), int(t_black_contrast_peaks.shape[0])],
                   'Total peak': [sci_num(mean_peak), sci_num(std_peak), int(t_contrast_peaks.shape[0])],
                   'Intersection bright': [sci_num(mean_bright_intersection), sci_num(std_bright_intersection), int(t_bright_contrast_intersection.shape[0])],
                   'Intersection black': [sci_num(mean_black_intersection), sci_num(std_black_intersection), int(t_black_contrast_intersection.shape[0])],
                   'Total Intersection': [sci_num(mean_intersection), sci_num(std_intersection), int(t_contrast_intersection.shape[0])],
                   'Prominence bright': [sci_num(mean_bright_prom), sci_num(std_bright_prom), int(t_bright_contrast_proms.shape[0])],
                   'Prominence black': [sci_num(mean_black_prom), sci_num(std_black_prom), int(t_black_contrast_proms.shape[0])],
                   'Total Prominence': [sci_num(mean_prom), sci_num(std_prom), int(t_contrast_proms.shape[0])]
                   }
        else:
            list_data = [t_bright_contrast_peaks, t_bright_contrast_intersection,
                         t_black_contrast_peaks, t_black_contrast_intersection,
                         t_contrast_peaks, t_contrast_intersection]

            title = ['Peak bright', 'Intersection bright',
                     'Peak black', 'Intersection black',
                     'Total peak', 'Total Intersection']

            dic = {'Peak bright': [sci_num(mean_bright_peak), sci_num(std_bright_peak),
                                   int(t_bright_contrast_peaks.shape[0])],
                   'Peak black': [sci_num(mean_black_peak), sci_num(std_black_peak),
                                  int(t_black_contrast_peaks.shape[0])],
                   'Total peak': [sci_num(mean_peak), sci_num(std_peak), int(t_contrast_peaks.shape[0])],
                   'Intersection bright': [sci_num(mean_bright_intersection), sci_num(std_bright_intersection),
                                           int(t_bright_contrast_intersection.shape[0])],
                   'Intersection black': [sci_num(mean_black_intersection), sci_num(std_black_intersection),
                                          int(t_black_contrast_intersection.shape[0])],
                   'Total Intersection': [sci_num(mean_intersection), sci_num(std_intersection),
                                          int(t_contrast_intersection.shape[0])]
                    }

        num_gmm_mean = None
        list_means = []
        list_stds = []
        list_weights = []
        list_keys = []
        list_num_GMM = []
        list_AIC = []
        list_BIC = []
        if Flag_GMM_fit:
            for data, key in zip(list_data, title):
                try:
                    means, stdevs, weights, AIC, BIC = self.GMM(np.abs(data), max_n_components)
                    list_num_GMM.append(len(means))
                    list_means.append(means)
                    list_stds.append(stdevs)
                    list_weights.append(weights)
                    list_AIC.append(AIC)
                    list_BIC.append(BIC)
                    list_keys.append(key)


                except:
                    print('---GMM did not extract for ' + key, '---')
                    list_means.append(None)
                    list_stds.append(None)
                    list_weights.append(None)
                    list_AIC.append(None)
                    list_BIC.append(None)
                    list_keys.append(key)
                    list_num_GMM.append(None)

            list_num_GMM_temp = [x for x in list_num_GMM if x is not None]
            max_num_gmm_mean = np.max(list_num_GMM_temp)

            for means, stdevs, weights, key in zip(list_means, list_stds, list_weights, list_keys):
                if means is not None:
                    if isinstance(means, list):
                        if len(means) == max_num_gmm_mean:
                            for m_, s_, w_ in zip(means, stdevs, weights):
                                dic[key].append(m_)
                                dic[key].append(s_)
                                dic[key].append(w_)
                        else:
                            diff_ = max_num_gmm_mean - len(means)
                            for m_, s_, w_ in zip(means, stdevs, weights):
                                dic[key].append(m_)
                                dic[key].append(s_)
                                dic[key].append(w_)

                            for i_ in range(diff_):
                                dic[key].append(None)
                                dic[key].append(None)
                                dic[key].append(None)
                    else:
                        diff_ = max_num_gmm_mean

                        for i_ in range(diff_):
                            dic[key].append(None)
                            dic[key].append(None)
                            dic[key].append(None)

                else:
                    for i_ in range(max_num_gmm_mean):
                        dic[key].append(None)
                        dic[key].append(None)
                        dic[key].append(None)

            index_ = ['Mean', 'Std', '#Particles']
            for i_ in range(max_num_gmm_mean):
                index_.append('GMM_mean' + str(i_+1))
                index_.append('GMM_std' + str(i_+1))
                index_.append('GMM_weight' + str(i_+1))
            df = pd.DataFrame(dic, index=index_)

        else:
            df = pd.DataFrame(dic, index=['Mean', 'Std', '#Particles'])

        return df, list_data, title

    def GMM(self, data, max_n_components):
        data = data.reshape(-1, 1)
        nan_array = np.isnan(data)
        not_nan_array = ~ nan_array
        data = data[not_nan_array]
        data = data.reshape(-1, 1)

        N = np.arange(1, max_n_components+1)
        models = [None for i in range(len(N))]
        for i in range(len(N)):
            try:
                models[i] = GaussianMixture(N[i], covariance_type='full', reg_covar=1e-10).fit(data)
            except:
                models[i] = None

        # compute the AIC and the BIC
        AIC = []
        BIC = []
        for m in models:
            if m is not None:
                AIC.append(m.aic(data))
                BIC.append(m.bic(data))
            else:
                AIC.append(np.nan)
                BIC.append(np.nan)

        AIC = [m.aic(data) for m in models if m is not None]
        BIC = [m.bic(data) for m in models if m is not None]

        try:
            num_mixture = np.min([np.nanargmin(AIC), np.nanargmin(BIC)])
            M_best = models[num_mixture]

            means = M_best.means_
            stdevs = np.sqrt(M_best.covariances_)
            weights = M_best.weights_

            means_ = means.tolist()
            stdevs_ = stdevs.tolist()
            weights_ = weights.tolist()
            sci_num = lambda x: "{:.2e}".format(x)

            means = [sci_num(m_[0]) for m_ in means_]
            stdevs = [sci_num(s_[0][0]) for s_ in stdevs_]
            weights = [sci_num(w_) for w_ in weights_]

            return means, stdevs, weights, AIC, BIC
        except:
            print('---Data is not enough for GMM!---')
            return np.nan, np.nan, np.nan, np.nan, np.nan

    def frameNumber_longerProfile_2_Vshape(self, batch_size, video_shape_0, particle_frame_number, V_smooth_follow):
        BS2 = 2 * batch_size
        start_FN_earlyV = particle_frame_number[0]
        end_FN_earlyV = particle_frame_number[-1]
        end_extended_DRA = np.min([video_shape_0, end_FN_earlyV + BS2])
        start_extended_DRA = np.max([0, start_FN_earlyV - BS2])

        tprofile_frameNo_DRA_longer = []
        tprofile_frameNo_DRA_longer.append(list(range(start_extended_DRA, start_FN_earlyV, 1)))
        tprofile_frameNo_DRA_longer.append(particle_frame_number)

        tprofile_frameNo_DRA_longer.append(list(range(end_FN_earlyV, end_extended_DRA, 1)))

        tprofile_frameNo_DRA_longer = [item for sublist in tprofile_frameNo_DRA_longer for item in sublist]

        if len(tprofile_frameNo_DRA_longer) != V_smooth_follow.shape[0]:
            pass

        return tprofile_frameNo_DRA_longer, start_FN_earlyV, end_FN_earlyV

    def plot_contrast_extraction(self, particles, batch_size, video_frame_num, MinPeakWidth,
                                 MinPeakProminence, pixel_size, particles_num='#0', save_path=None):

        if type(particles) is list:
            particles = protein_trajectories_list2dic(particles)

        if type(particles) is dict:
            for key, item in particles.items():
                if key == particles_num:
                    intensity_horizontal = item['intensity_horizontal']
                    intensity_vertical = item['intensity_vertical']
                    center_int = item['center_int']
                    center_int_flow = item['center_int_flow']
                    frame_number = item['frame_number']
                    sigma = item['sigma']
                    x_center = item['x_center']
                    y_center = item['y_center']
                    particle_ID = item['particle_ID']

                    num_parameters = len(item)

                    if num_parameters == 21:
                        fit_intensity = item['fit_intensity']
                        fit_x = item['fit_x'] * pixel_size
                        fit_y = item['fit_y'] * pixel_size
                        fit_X_sigma = item['fit_X_sigma'] * pixel_size
                        fit_Y_sigma = item['fit_Y_sigma'] * pixel_size
                        fit_Bias = item['fit_Bias']
                        fit_intensity_error = item['fit_intensity_error']
                        fit_x_error = item['fit_x_error'] * pixel_size
                        fit_y_error = item['fit_y_error'] * pixel_size
                        fit_X_sigma_error = item['fit_X_sigma_error'] * pixel_size
                        fit_Y_sigma_error = item['fit_Y_sigma_error'] * pixel_size
                        fit_Bias_error = item['fit_Bias_error']

                    else:
                        fit_intensity = None
                        fit_X_sigma = None
                        fit_Y_sigma = None

                    self.t_particle_localization_flag = True
                    properties, fit_properties = self.data_handling(center_int, center_int_flow, '', sigma,
                                                                    num_parameters, fit_intensity, fit_X_sigma,
                                                                    fit_Y_sigma, frame_number, batch_size,
                                                                    video_frame_num, MinPeakWidth=MinPeakWidth,
                                                                    MinPeakProminence=MinPeakProminence)

                    if properties is not None:

                        self.axs.plot(properties[5], properties[1], 'C0--', linewidth=5, zorder=0)
                        self.axs.plot(frame_number, properties[0], 'b', linewidth=5, zorder=0)
                        self.axs.plot(properties[2][1], properties[2][0], 'gv', label='Peak', markersize=7, zorder=4)
                        self.axs.axvline(properties[6], linestyle='--', zorder=5)
                        self.axs.axvline(properties[7], linestyle='--', zorder=6)
                        if math.isnan(properties[4][1]) and math.isnan(properties[4][0]):
                            flag_skip_prominence = True
                        else:
                            flag_skip_prominence = False
                            if properties[2][0] >= 0:
                                self.axs.vlines(x=properties[4][1], ymin=properties[4][0] - properties[4][3],
                                                ymax=properties[4][0], color="C1", label='Prominence')
                            else:
                                self.axs.vlines(x=properties[4][1], ymin=properties[4][0] + properties[4][3],
                                                ymax=properties[4][0], color="C1", label='Prominence')
                        self.axs.set_ylim(np.min(properties[1]) + 0.5*np.min(properties[1]), np.max(properties[1]) + 0.5*np.max(properties[1]))
                        self.axs.grid()
                        self.axs.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
                        if flag_skip_prominence:
                            handles, labels = self.axs.get_legend_handles_labels()

                            handles = [handles[2], handles[0], handles[1], handles[3]]
                            labels = [labels[2], labels[0], labels[1], labels[3]]
                        else:
                            handles, labels = self.axs.get_legend_handles_labels()

                            handles = [handles[2], handles[0], handles[1], handles[3], handles[4]]
                            labels = [labels[2], labels[0], labels[1], labels[3], labels[4]]
                        self.axs.legend(handles, labels)
                        if save_path is not None:
                            print('plot saved!')
                            self.fig.savefig(save_path, dpi=500)

                        plt.show()
                    else:
                        raise Exception('---Data can not extract!---')

    def plot_histogram(self, bins=None, upper_limitation=1, lower_limitation=-1,
                       step_range=1e-7, face='g', edge='y', Flag_GMM_fit=True, max_n_components=3, imgSizex=20,
                       imgSizey=20, font_size=12, scale=1e1, external_GMM=False):
        """
        This method plots histograms for different contrast extraction methods for black PSFs, white PSFs and all together.

        Parameters
        ----------
        bins: int
            Number of histogram bins.

        upper_limitation: float
            The upper limit for trimming histogram.

        lower_limitation: float
            The lower limit for trimming histogram.

        step_range: float
            The resolution that is used for GMM plotting.

        face: str
            Face color of the histogram.

        edge: str
            Edge color of the histogram.

        Flag_GMM_fit: bool
            Activate/Deactivate GMM.

        max_n_components: int
            The maximum number of components that GMM used for AIC and BIC tests. This helps to find an optimum number of the mixture.

        imgSizex: int
            The width of the histogram figure.

        imgSizey: int
            The height of the histogram figure.

        font_size: float
            The font size of the text in the table information.

        scale: float
            This value multiplies the full range for x-axis plotting.

        external_GMM: bool
            This flag modifies GMM's visualization. Only the external border is visible if it is set to True.
        """

        df, list_data, title = self.extract_hist_information(con_intersections=self.t_contrast_intersection,
                                                             con_peaks=self.t_contrast_peaks,
                                                             con_proms=self.t_contrast_Prominence,
                                                             upper_limitation=upper_limitation,
                                                             lower_limitation=lower_limitation,
                                                             max_n_components=max_n_components,
                                                             Flag_GMM_fit=Flag_GMM_fit)

        fig = plt.figure(figsize=(imgSizex, imgSizey))
        outer = gridspec.GridSpec(2, 1, wspace=0.3, hspace=0.2)

        inner = gridspec.GridSpecFromSubplotSpec(3, 3,
                                                 subplot_spec=outer[0], wspace=0.3, hspace=0.7)

        for p_index in range(len(list_data)):
            if Flag_GMM_fit:
                try:
                    key = title[p_index]
                    index_ = df.index.to_list()
                    pdfs = []
                    means = []
                    stdevs = []
                    weights =[]
                    for idx_ in index_:
                        if 'GMM_mean' in idx_:
                            means.append(df[key][idx_])
                        elif 'GMM_std' in idx_:
                            stdevs.append(df[key][idx_])
                        elif 'GMM_weight' in idx_:
                            weights.append(df[key][idx_])

                    d_ = np.abs(list_data[p_index])
                    min_data = np.min(d_) - 0.5 * np.min(d_)
                    max_data = np.max(d_) + 0.5 * np.max(d_)

                    x = np.arange(min_data, max_data, step_range)

                    pdfs = [float(p) * ss.norm.pdf(x, float(mu), float(sd)) for mu, sd, p in zip(means, stdevs, weights)
                            if mu is not None and sd is not None and p is not None]
                    density = np.sum(np.array(pdfs), axis=0)
                    ax = plt.Subplot(fig, inner[p_index])
                    ax.hist(d_, bins=bins, fc=face, ec=edge, density=True)
                    ax.axis(xmin=0, xmax=upper_limitation * scale)
                    if external_GMM:
                        ax.plot(x.ravel(), density.ravel())
                    else:
                        for j_ in range(len(weights)):
                            if weights[j_] is not None:
                                ax.plot(x.ravel(), float(weights[j_]) * stats.norm.pdf(x.ravel(), float(means[j_]), float(stdevs[j_])).ravel())

                    ax.set_ylabel('Density')
                    ax.set_title(title[p_index])
                    ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
                    fig.add_subplot(ax)

                except:
                    d_ = np.abs(list_data[p_index])

                    ax = plt.Subplot(fig, inner[p_index])
                    ax.hist(d_, bins=bins, fc=face, ec=edge, density=False)
                    ax.axis(xmin=0, xmax=upper_limitation * scale)
                    ax.set_ylabel('#Counts')
                    ax.set_title(title[p_index])
                    ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
                    fig.add_subplot(ax)

            else:
                d_ = np.abs(list_data[p_index])

                ax = plt.Subplot(fig, inner[p_index])
                ax.hist(d_, bins=bins, fc=face, ec=edge, density=False)
                ax.axis(xmin=0, xmax=upper_limitation * scale)
                ax.set_ylabel('#Counts')
                ax.set_title(title[p_index])
                ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
                fig.add_subplot(ax)

        ax2 = plt.Subplot(fig, outer[1])
        font_size = font_size
        bbox = [0, 0, 1, 1]
        ax2.axis('off')
        colors = plt.cm.BuPu(np.linspace(0, 0.5, df.shape[0]))
        cell_colors = [[colors[i_]] * df.shape[1] for i_ in range(df.shape[0])]
        mpl_table = ax2.table(cellText=df.values, rowLabels=df.index, bbox=bbox, colLabels=df.columns,  rowColours=colors,
                              cellColours=cell_colors)
        mpl_table.auto_set_font_size(False)
        mpl_table.set_fontsize(font_size)
        fig.add_subplot(ax2)

        plt.show()

    def plot_fit_histogram(self, bins=None, upper_limitation=1, lower_limitation=-1, step_range=1e-7, face='g',
                           edge='y', Flag_GMM_fit=True, max_n_components=3, imgSizex=20, imgSizey=20, font_size=12,
                           scale=1e1, external_GMM=False):
        """
        This method plots histograms for 2D Gaussian fitting contrast for black PSFs, white PSFs and all together.

        Parameters
        ----------
        bins: int
           Number of histogram bins.

        upper_limitation: float
           The upper limit for trimming histogram.

        lower_limitation: float
           The lower limit for trimming histogram.

        step_range: float
           The resolution that is used for GMM plotting.

        face: str
           Face color of the histogram.

        edge: str
           Edge color of the histogram.

        Flag_GMM_fit: bool
           Activate/Deactivate GMM.

        max_n_components: int
           The maximum number of components that GMM used for AIC and BIC tests. This helps to find an optimum number of the mixture.

        imgSizex: int
           The width of the histogram figure.

        imgSizey: int
           The height of the histogram figure.

        font_size: float
           The font size of the text in the table information.

        scale: float
                   This value multiplies the full range for x-axis plotting.

        external_GMM: bool
           This flag modifies GMM's visualization. Only the external border is visible if it is set to True.

        """

        df, list_data, title = self.extract_hist_information(con_intersections=self.t_contrast_fit_intersection,
                                                             con_peaks=self.t_contrast_fit_peaks,
                                                             con_proms=None,
                                                             upper_limitation=upper_limitation,
                                                             lower_limitation=lower_limitation,
                                                             max_n_components=max_n_components,
                                                             Flag_GMM_fit=Flag_GMM_fit)

        fig = plt.figure(figsize=(imgSizex, imgSizey))
        outer = gridspec.GridSpec(2, 1, wspace=0.3, hspace=0.2)

        inner = gridspec.GridSpecFromSubplotSpec(3, 3,
                                                 subplot_spec=outer[0], wspace=0.3, hspace=0.7)

        for p_index in range(len(list_data)):
            if Flag_GMM_fit:
                try:
                    key = title[p_index]
                    index_ = df.index.to_list()
                    pdfs = []
                    means = []
                    stdevs = []
                    weights = []
                    for idx_ in index_:
                        if 'GMM_mean' in idx_:
                            means.append(df[key][idx_])
                        elif 'GMM_std' in idx_:
                            stdevs.append(df[key][idx_])
                        elif 'GMM_weight' in idx_:
                            weights.append(df[key][idx_])

                    d_ = np.abs(list_data[p_index])
                    min_data = np.min(d_) - 0.5 * np.min(d_)
                    max_data = np.max(d_) + 0.5 * np.max(d_)

                    x = np.arange(min_data, max_data, step_range)

                    pdfs = [float(p) * ss.norm.pdf(x, float(mu), float(sd)) for mu, sd, p in zip(means, stdevs, weights)
                            if mu is not None and sd is not None and p is not None]
                    density = np.sum(np.array(pdfs), axis=0)
                    ax = plt.Subplot(fig, inner[p_index])
                    ax.hist(d_, bins=bins, fc=face, ec=edge, density=True)
                    ax.axis(xmin=0, xmax=upper_limitation * scale)
                    if external_GMM:
                        ax.plot(x.ravel(), density.ravel())
                    else:
                        for j_ in range(len(weights)):
                            if weights[j_] is not None:
                                ax.plot(x.ravel(), float(weights[j_]) * stats.norm.pdf(x.ravel(), float(means[j_]),
                                                                                       float(stdevs[j_])).ravel())

                    ax.set_ylabel('Density')
                    ax.set_title(title[p_index])
                    ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
                    fig.add_subplot(ax)

                except:
                    d_ = np.abs(list_data[p_index])

                    ax = plt.Subplot(fig, inner[p_index])
                    ax.hist(d_, bins=bins, fc=face, ec=edge, density=False)
                    ax.axis(xmin=0, xmax=upper_limitation * scale)
                    ax.set_ylabel('#Counts')
                    ax.set_title(title[p_index])
                    ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
                    fig.add_subplot(ax)

            else:
                d_ = np.abs(list_data[p_index])

                ax = plt.Subplot(fig, inner[p_index])
                ax.hist(d_, bins=bins, fc=face, ec=edge, density=False)
                ax.axis(xmin=0, xmax=upper_limitation * scale)
                ax.set_ylabel('#Counts')
                ax.set_title(title[p_index])
                ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
                fig.add_subplot(ax)

        ax2 = plt.Subplot(fig, outer[1])
        font_size = font_size
        bbox = [0, 0, 1, 1]
        ax2.axis('off')
        mpl_table = ax2.table(cellText=df.values, rowLabels=df.index, bbox=bbox, colLabels=df.columns)
        mpl_table.auto_set_font_size(False)
        mpl_table.set_fontsize(font_size)
        fig.add_subplot(ax2)

        plt.show()

    def plot_fit_sigma(self):
        mean_sigmX = np.mean(self.t_iPSFCentroidSigmas_fit_x)
        mean_sigmY = np.mean(self.t_iPSFCentroidSigmas_fit_y)
        if mean_sigmY >= mean_sigmX:
            ratio_sigma = mean_sigmY / mean_sigmX
        else:
            ratio_sigma = mean_sigmX / mean_sigmY

        plt.figure()
        plt.hist(self.t_iPSFCentroidSigmas_fit_x)
        plt.hist(self.t_iPSFCentroidSigmas_fit_y)
        plt.xlabel('sigma(nm)')
        plt.ylabel('#Counts')
        plt.legend('Fitted sigma_x', 'Fitted sigma_y')
        plt.title('Ratio mean of Sigma_Max/Sigma_Min=' + str(ratio_sigma))
        plt.show()

    def plot_localization_heatmap(self, pixel_size, unit, flag_in_time=False, time_delay=0.1, dir_name=None):
        """
        This method generates a particle localization heatmap. Every disk's size represents the movement of each particle during tracking.

        Parameters
        ----------
        pixel_size: float
            camera pixel size

        unit: str
            The axis unit.

        flag_in_time: bool
            In the case of True, show binding and unbinding events in time.

        time_delay: float
            Define the time delay between binding and unbinding events frames. This only works when `flag_in_time` is set to True.

        dir_name: str
            You can save time slap frames if you specify a save path.
        """
        bubble_size_bright_ = np.maximum(self.t_std_y_center_bright, self.t_std_x_center_bright)
        bubble_size_dark_ = np.maximum(self.t_std_y_center_dark, self.t_std_x_center_dark)

        print('Number of bright particles {}'.format(bubble_size_bright_.shape[0]))
        print('Number of dark particles {}'.format(bubble_size_dark_.shape[0]))

        bubble_size_bright = []
        for s_b in bubble_size_bright_:
            if s_b == 0:
                bubble_size_bright.append(s_b + 1e-6)
            else:
                bubble_size_bright.append(s_b)

        bubble_size_dark = []
        for s_d in bubble_size_dark_:

            if s_d == 0:
                bubble_size_dark.append(s_d + 1e-6)
            else:
                bubble_size_dark.append(s_d)

        dic_dark = {'x': np.multiply(pixel_size, self.t_mean_x_center_dark),
                    'y': np.multiply(pixel_size, self.t_mean_y_center_dark),
                    'bubble_size': np.multiply(pixel_size, bubble_size_dark),
                    'particle': np.asarray(self.t_particle_ID_dark),
                    'frame': self.t_particle_frame_dark
                    }

        df_dark = pd.DataFrame.from_dict(dic_dark)

        dic_bright = {'x': np.multiply(pixel_size, self.t_mean_x_center_bright),
                      'y': np.multiply(pixel_size, self.t_mean_y_center_bright),
                      'bubble_size': np.multiply(pixel_size, bubble_size_bright),
                      'particle': np.asarray(self.t_particle_ID_bright),
                      'frame': self.t_particle_frame_bright
                      }


        df_bright = pd.DataFrame.from_dict(dic_bright)

        df_bright_sort = df_bright.sort_values('frame')
        df_dark_sort = df_dark.sort_values('frame')
        df_bright_sort = df_bright_sort.reset_index()
        df_bright_sort.particle = df_bright_sort.index

        if flag_in_time:
            plot_bright_dark_psf_inTime(df_bright_sort, df_dark_sort, time_delay, dir_name)
        else:
            plot_bright_dark_psf(df_bright_sort, df_dark_sort, unit)

    def plot_linking_length_histogram(self, bins=None):

        fig, axs = plt.subplots(1, 3, sharey=True, tight_layout=True)
        axs[0].hist(self.t_linking_len_bright, bins=bins)
        axs[0].set_title('bright PSF linking length')

        axs[1].hist(self.t_linking_len_dark, bins=bins)
        axs[1].set_title('dark PSF linking length')

        total_len = self.t_linking_len_bright + self.t_linking_len_dark
        axs[2].hist(total_len, bins=bins)
        axs[2].set_title('total PSF linking length')

        plt.show()

    def save_hist_data(self, dirName, name, upper_limitation=1, lower_limitation=-1, Flag_GMM_fit=True, max_n_components=3):
        """
        This function save the histogram data with HDF5 format.

        Parameters
        ----------
        dirName: str
            Path for saving data.

        name: str
            Name that use for saving data.

        upper_limitation: float
            The upper limit for trimming histogram.

        lower_limitation: float
            The lower limit for trimming histogram.

        Flag_GMM_fit: bool
            Activate/Deactivate GMM.

        max_n_components: int
            The maximum number of components that GMM used for AIC and BIC tests. This helps to find an optimum number of the mixture.

        """
        df, list_data, title = self.extract_hist_information(con_intersections=self.t_contrast_intersection,
                                                             con_peaks=self.t_contrast_peaks,
                                                             con_proms=self.t_contrast_Prominence,
                                                             upper_limitation=upper_limitation,
                                                             lower_limitation=lower_limitation,
                                                             max_n_components=max_n_components,
                                                             Flag_GMM_fit=Flag_GMM_fit)
        dic_data = {}
        for d_, t_ in zip(list_data, title):
            dic_data[t_] = d_

        read_write_data.save_dic_to_hdf5(dic_data=dic_data, path=dirName, name=name)

        dic_ = {'con_intersections': self.t_contrast_intersection,
                'con_peaks': self.t_contrast_peaks,
                'con_proms': self.t_contrast_Prominence,
                'len_linking_intersections': self.t_len_linking}

        read_write_data.save_dic_to_hdf5(dic_data=dic_, path=dirName, name=name+'_data_hist_without_trim')

    def localization_data(self):
        """
        This method generates a particle localization heatmap. Every disk's size represents the movement of each particle during tracking.

        Parameters
        ----------
        pixel_size: float
            camera pixel size

        unit: str
            unit of axises
        """
        bubble_size_bright_ = np.maximum(self.t_std_y_center_bright, self.t_std_x_center_bright)
        bubble_size_dark_ = np.maximum(self.t_std_y_center_dark, self.t_std_x_center_dark)



        dic_dark = {'x': self.t_mean_x_center_dark,
                    'y': self.t_mean_y_center_dark,
                    'bubble_size': bubble_size_dark_,
                    'particle': np.asarray(self.t_particle_ID_dark),
                    'frame': self.t_particle_frame_dark
                    }

        df_dark = pd.DataFrame.from_dict(dic_dark)

        dic_bright = {'x': self.t_mean_x_center_bright,
                      'y': self.t_mean_y_center_bright,
                      'bubble_size': bubble_size_bright_,
                      'particle': np.asarray(self.t_particle_ID_bright),
                      'frame': self.t_particle_frame_bright
                      }

        df_bright = pd.DataFrame.from_dict(dic_bright)

        return df_bright, df_dark







"""
__author__ = "Houman Mirzaalian D."
__license__ = "GPL"
__email__ = "hmirzaa@mpl.mpg.de
"""

from base64 import b64encode, b64decode
import json
import re
import pandas as pd

from PySide6.QtGui import *
from PySide6.QtWidgets import *
from PySide6.QtCore import *

CURSOR_DEFAULT = Qt.ArrowCursor
CURSOR_POINT = Qt.PointingHandCursor
CURSOR_DRAW = Qt.CrossCursor
CURSOR_MOVE = Qt.ClosedHandCursor
CURSOR_GRAB = Qt.OpenHandCursor


class LineWidthDialog(QDialog):
    '''A window with a slider to set line width of annotations
    '''

    def __init__(self, allshapes=None, scene=None, parent=None):
        super(LineWidthDialog, self).__init__(parent)
        self.setWindowTitle("Line width editor")
        self.scene = scene
        self.allshapes = allshapes
        self.lwidth = allshapes[0].point_size

        self.form = QGridLayout(self)

        self.form.addWidget(QLabel("Please set line width"), 0, 0)

        self.slider = QSlider(Qt.Horizontal)
        self.slider.setTickPosition(QSlider.TicksBothSides)

        self.slider.setMinimum(0)
        self.slider.setMaximum(10)
        self.slider.setValue(self.lwidth)
        self.form.addWidget(self.slider, 1, 0)
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, Qt.Horizontal)

        self.form.addWidget(self.buttonBox, 2, 0)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.accepted.connect(self.extractInputs)

    def extractInputs(self):
        self.size = self.slider.value()
        if self.size != self.lwidth:
            self.scene.point_size = self.size
            for shape in self.allshapes:
                shape.point_size = self.size


class EpsilonSliderDialog(QDialog):
    '''A window with a slider to set attraction epsilon, i.e. pull the cursor
    to the first point of the shape if epsilon close
    '''

    def __init__(self, scene=None, parent=None):
        super(EpsilonSliderDialog, self).__init__(parent)
        self.setWindowTitle("Attraction epsilon editor")
        self.scene = scene

        self.epsilon = self.scene.epsilon

        self.form = QGridLayout(self)

        self.form.addWidget(QLabel("Please set epsilon"), 0, 0)

        self.slider = QSlider(Qt.Horizontal)
        self.slider.setTickPosition(QSlider.TicksBothSides)

        self.slider.setMinimum(0)
        self.slider.setMaximum(2 * self.epsilon)
        self.slider.setValue(self.epsilon)
        self.form.addWidget(self.slider, 1, 0)
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, Qt.Horizontal)

        self.form.addWidget(self.buttonBox, 2, 0)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.accepted.connect(self.extractInputs)

    def extractInputs(self):
        self.newepsilon = self.slider.value()
        if self.newepsilon != self.epsilon:
            self.scene.epsilon = self.newepsilon


class PropertiesWindow(QDialog):
    '''A window that pops up on right click (see mousepressevent in qscene).
    Lets the user reassign an annotated ("closed") shape to a different label class.
    '''

    def __init__(self, shape=None, all_labels=[None], scene=None, parent=None):
        super(PropertiesWindow, self).__init__(parent)
        self.shape = shape
        self.points = self.shape.points
        self.objtype = self.shape.objtype
        self.label = self.shape.label
        self.all_labels = all_labels
        self.setWindowTitle("Properties editor")
        self.label_names = [label.name for label in all_labels if label is not None]
        self.editable_status = self.shape.editable

        self.form = QGridLayout(self)

        self.form.addWidget(QLabel("Object properties"), 0, 0)

        self.qbox = QComboBox(self)
        self.qbox.addItem(self.label)
        [self.qbox.addItem(label) for label in self.label_names if label != self.label]

        self.form.addWidget(QLabel("Type:"), 1, 0)
        self.form.addWidget(QLabel(self.objtype), 1, 1)
        self.form.addWidget(QLabel("Label:"), 2, 0)
        self.form.addWidget(self.qbox, 2, 1)

        copybutton = QPushButton('Copy', self)
        copybutton.clicked.connect(scene.copySelected)

        self.editbutton = QPushButton('Set editable', self)
        self.editbutton.setCheckable(True)
        self.editbutton.clicked.connect(self.updateCheckable)
        self.editableColor()

        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, Qt.Horizontal)
        self.buttonBox.addButton(copybutton, QDialogButtonBox.AcceptRole)
        self.buttonBox.addButton(self.editbutton, QDialogButtonBox.ActionRole)
        self.form.addWidget(self.buttonBox, 3, 0)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.accepted.connect(self.extractInputs)

    def editableColor(self):
        if self.editable_status:
            color = QColor(255, 0, 0)
        else:
            color = QColor(255, 255, 255)

        palette = self.editbutton.palette()
        role = self.editbutton.backgroundRole()
        palette.setColor(role, color)
        self.editbutton.setPalette(palette)
        self.editbutton.setAutoFillBackground(True)
        return

    def updateCheckable(self, checked=False):
        if checked and not self.shape.editable:
            self.editable_status = True
        else:
            self.editable_status = False
        self.editableColor()
        return

    def extractInputs(self):
        self.shape.editable = self.editable_status
        oldlabel = self.shape.label
        newlabel = self.qbox.currentText()
        if oldlabel != newlabel:
            if oldlabel in self.label_names:
                self.all_labels[self.label_names.index(oldlabel)].untieShape(self.shape)
                newlabelclass = self.all_labels[self.label_names.index(newlabel)]
                newlabelclass.assignObject(self.shape)
        return


class LabelDialog(QDialog):
    '''Label editor form. Given # of new classes to initiate (nlabels) pops up
    a form with a line editor to name a class and a button to set label color.
    If nlabel=0, the form Features only already initialized labels for editing.
    '''

    def __init__(self, nlabels=1, prelabels=None, parent=None):
        super(LabelDialog, self).__init__(parent)
        self.names = []
        self.colors = [None for i in range(nlabels)]
        self.prelabels = prelabels
        self.setWindowTitle("Initialize labels")
        self.lineedits = []

        self.form = QGridLayout(self)
        self.form.addWidget(QLabel("Please give description of labels below:"), 0, 0)
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, Qt.Horizontal)

        if prelabels is not None:
            self.prenames = [label.name for label in prelabels]
            self.precolors = [label.fillColor for label in prelabels]
            nprelabels = len(prelabels)
            self.colors = self.colors + nprelabels * [None]
            [self.addLabelInfo(index, prelabdata=[self.prenames[index], self.precolors[index]]) for index in
             range(nprelabels)]
            [self.addLabelInfo(index) for index in range(nprelabels, nprelabels + nlabels)]
            self.form.addWidget(self.buttonBox, 2 * (nprelabels + nlabels) + 1, 0)
        else:
            [self.addLabelInfo(index) for index in range(nlabels)]
            self.form.addWidget(self.buttonBox, 2 * nlabels + 1, 0)

        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.accepted.connect(self.extractInputs)
        self.rejected.connect(self.clearInputs)

    def fillcolornone(self):
        if None in self.colors:
            ind = self.colors.index(None)
            fillcolor1 = int(255 * ind / len(self.colors))
            fillcolor2 = int(255 / len(self.colors))
            fillcolor3 = 255
            self.colors[self.colors.index(None)] = QColor(fillcolor1, fillcolor2, fillcolor3)
            self.fillcolornone()
        return

    def fillemptynames(self):
        indices = [i for i, name in enumerate(self.names) if name == '']
        for ind in indices:
            self.names[ind] = 'class' + str(ind + 1)
        return

    def extractInputs(self):
        self.names = [lineedit.text() for lineedit in self.lineedits]
        self.fillemptynames()
        self.fillcolornone()
        return

    def clearInputs(self):
        self.names = []
        self.colors = [None for i in range(len(self.colors))]
        if self.prelabels is not None:
            self.names = self.prenames
            self.colors = self.precolors
        return

    def addLabelInfo(self, labelindex, prelabdata=2 * [None]):
        lineEdit = QLineEdit(self)
        button = QPushButton('Select color', self)
        labelname = QLabel('Label {} name'.format(labelindex + 1))
        labelcolor = QLabel('Label {} color'.format(labelindex + 1))

        if prelabdata[0] is not None:
            lineEdit.setText("{}".format(prelabdata[0]))
            color = prelabdata[1]
            self.colors[labelindex] = color
            palette = button.palette()
            role = button.backgroundRole()
            palette.setColor(role, color)
            button.setPalette(palette)
            button.setAutoFillBackground(True)

        button.clicked.connect(lambda: self.selectColor(labelindex))

        self.lineedits.append(lineEdit)
        self.form.addWidget(labelname, 2 * labelindex + 1, 0)
        self.form.addWidget(lineEdit, 2 * labelindex + 1, 1)
        self.form.addWidget(labelcolor, 2 * labelindex + 2, 0)
        self.form.addWidget(button, 2 * labelindex + 2, 1)

    def selectColor(self, labelindex):
        dialogue = QColorDialog()
        dialogue.exec()
        color = dialogue.selectedColor()
        self.colors[labelindex] = color

        button = QPushButton('Select color', self)
        palette = button.palette()
        role = button.backgroundRole()
        palette.setColor(role, color)
        button.setPalette(palette)
        button.setAutoFillBackground(True)
        button.clicked.connect(lambda: self.selectColor(labelindex))
        self.form.addWidget(button, 2 * labelindex + 2, 1)
        return


class ToolButton(QToolButton):
    '''Overwrite QToolButton to ensure all buttons are of same size
    '''
    minSize = (80, 80)

    def minimumSizeHint(self):
        ms = super(ToolButton, self).minimumSizeHint()
        w1, h1 = ms.width(), ms.height()
        w2, h2 = self.minSize
        ToolButton.minSize = max(w1, w2), max(h1, h2)
        return QSize(*ToolButton.minSize)


def process(filename, default=None):
    '''Return bytes for an image given its path
    '''
    try:
        with open(filename, 'rb') as f:
            return f.read()
    except:
        return default


def newAction(parent, text, slot=None, shortcut=None, icon=None,
              tip=None, checkable=False, enabled=True):
    '''Initialize an action with flags as requested'''
    a = QAction(text, parent)

    if shortcut is not None:
        if isinstance(shortcut, (list, tuple)):
            a.setShortcuts(shortcut)
        else:
            a.setShortcut(shortcut)
    if tip is not None:
        a.setToolTip(tip)
    if slot is not None:
        a.triggered.connect(slot)
    if checkable:
        a.setCheckable(True)
    a.setEnabled(enabled)
    return a


def distance(delta):
    '''Squared euclidean distance as metric. Returns distance given an elementwise delta
    '''
    return (delta.x() ** 2 + delta.y() ** 2)


class LabelClass(object):
    '''Class to keep record of a label class characteristics with a method to
    assign shapes to it
    '''

    def __init__(self):
        self.polygons = []
        self.fillColor = None
        self.name = None

    def assignObject(self, obj):
        self.polygons.append(obj)
        obj.line_color = self.fillColor
        obj.label = self.name

    def untieShape(self, obj):
        self.polygons.pop(self.polygons.index(obj))


class Annotationscene(object):
    '''Class to store and save outputs as .csv and .json
    '''

    def __init__(self, filename=None):
        self.polygons = None
        self.imagePath = None
        self.imageData = None
        self.filename = None
        self.lineColor = None
        self.imsizes = None
        self.object_types = None
        self.labels = None
        self.savebytes = False

    def shapes_to_pandas(self):
        imsize, objects, types, labels = self.imsizes, self.shapes, self.object_types, self.labels
        df = pd.DataFrame(columns=['width', 'height', 'Object', 'X', 'Y'])
        for i, obj in enumerate(objects):
            X, Y = list(zip(*obj))
            df = df.append(pd.DataFrame(
                {'width': imsize[0], 'height': imsize[1], 'Object': i + 1, 'Type': types[i], 'Label': labels[i], 'X': X,
                 'Y': Y}), ignore_index=True)
        return df

    def save(self):

        self.shapes = [[(point.x(), point.y()) for point in poly] for poly in self.polygons]
        self.shapes_to_pandas().to_csv(re.search(re.compile('(.+?)(\.[^.]*$|$)'), self.filename).group(1) + '.csv',
                                       sep=',')
        if self.savebytes:
            self.imData = b64encode(self.imageData).decode('utf-8')

            try:
                with open(self.filename, 'w') as f:

                    json.dump({
                        'objects': self.shapes,
                        'bin_type': self.object_types,
                        'label': self.labels,
                        'width/height': self.imsizes,
                        'lineColor': self.lineColor,
                        'imagePath': self.imagePath,
                        'imageData': self.imData},
                        f, ensure_ascii=True, indent=2)
            except:
                pass
        else:
            try:
                with open(self.filename, 'w') as f:
                    json.dump({
                        'objects': self.shapes,
                        'bin_type': self.object_types,
                        'label': self.labels,
                        'width/height': self.imsizes,
                        'lineColor': self.lineColor,
                        'imagePath': self.imagePath},
                        f, ensure_ascii=True, indent=2)
            except:
                pass


class Shape(QGraphicsItem):
    '''The main class controlling shape's points, its color, highlight behavior
    '''
    line_color = QColor(0, 6, 255)
    select_line_color = QColor(255, 255, 255)
    vertex_fill_color = QColor(0, 255, 0, 255)
    hvertex_fill_color = QColor(255, 0, 0)
    point_size = 1.5
    hsize = 3.0

    def __init__(self, line_color=None, point_size=None, parent=None):
        super(Shape, self).__init__(parent)
        self.points = []
        self.selected = False
        self.painter = QPainter()
        self.hIndex = None
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.closed = False
        self.points_adjusted = None
        self.objtype = None
        self.label = None
        self.editable = False

        if line_color is not None:
            self.line_color = line_color

        if point_size is not None:
            self.point_size = point_size

    def addPoint(self, point):
        self.setSelected(True)
        if self.points and point == self.points[0]:
            self.closed = True
        else:
            self.points.append(point)

    def popPoint(self):
        if self.points:
            return self.points.pop()
        return None

    def paint(self, painter, option, widget):

        if self.points:
            self.prepareGeometryChange()
            color = self.select_line_color if self.selected else self.line_color
            pen = QPen(color)
            pen.setWidth(self.point_size / 2)
            painter.setPen(pen)
            path = self.shape()
            if self.closed == True:
                path.closeSubpath()
            painter.drawPath(path)
            vertex_path = QPainterPath()
            self.drawVertex(vertex_path, 0)
            [self.drawVertex(vertex_path, i) for i in range(len(self.points))]
            painter.drawPath(vertex_path)
            painter.fillPath(vertex_path, self.vertex_fill_color)

    def drawVertex(self, path, index):
        psize = self.point_size
        if index == self.hIndex:
            psize = self.hsize
        if self.hIndex is not None:
            self.vertex_fill_color = self.hvertex_fill_color
        else:
            self.vertex_fill_color = Shape.vertex_fill_color
        path.addEllipse(self.mapFromScene(self.points[index]), psize, psize)

    def shape(self):
        path = QPainterPath()
        polygon = self.mapFromScene(QPolygonF(self.points))
        path.addPolygon(polygon)
        return path

    def boundingRect(self):
        return self.shape().boundingRect()

    def moveBy(self, tomove, delta):
        if tomove == 'all':
            tomove = slice(0, len(self.points))
        else:
            tomove = slice(tomove, tomove + 1)
        self.points[tomove] = [point + delta for point in self.points[tomove]]

    def highlightVertex(self, index):
        self.hIndex = index

    def highlightClear(self):
        self.hIndex = None
        self.selected = False

    def __len__(self):
        return len(self.points)

    def __getitem__(self, index):
        return self.points[index]

    def __setitem__(self, index, value):
        self.points[index] = value


class SubQGraphicsScene(QGraphicsScene):
    '''Overwrite QGraphicsScene to prescribe actions to mouse events,
    collect annotated shapes and label classes, tracks which mode the program is in
    at any moment (drawing, navigating, moving)
    '''
    NAVIGATION, DRAWING, MOVING = 0, 1, 2
    POLYDRAWING, POLYREADY = 0, 1
    epsilon = 30.0

    def __init__(self, parent=None):
        super(SubQGraphicsScene, self).__init__(parent)
        self.mode = self.NAVIGATION
        self.QGitem = None
        self.polys = []
        self._cursor = CURSOR_DEFAULT
        self.overrideCursor(self._cursor)
        self.line = None
        self.lineColor = QColor(3, 252, 66)
        self.shapeColor = QColor(0, 6, 255)
        self.selectedVertex = None
        self.selectedShape = None
        self.polystatus = self.POLYDRAWING
        self.objtypes = []
        self.labelclasses = []
        self.labelmode = 0  # the default class
        self.initializeClass('default', QColor(0, 6, 255))  # initialize default class
        self.point_size = 1.5

    def drawing(self):
        return self.mode == self.DRAWING

    def navigating(self):
        return self.mode == self.NAVIGATION

    def moving(self):
        return self.mode == self.MOVING

    def polygon_not_finished(self):
        return self.polystatus == self.POLYDRAWING

    def polygonfinished(self):
        return self.polystatus == self.POLYREADY

    def vertexSelected(self):
        return self.selectedVertex is not None

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Delete:
            self.deleteSelected()
        if event.key() == Qt.Key_K:
            self.copySelected()
        if (event.key() == Qt.Key_Z) and (QApplication.keyboardModifiers() == Qt.ControlModifier):
            self.undoAction()

    def undoAction(self):
        if self.QGitem:
            if len(self.QGitem.points) > 1:
                self.QGitem.popPoint()
                self.line.points[0] = self.QGitem.points[-1]
                self.update()
            else:
                self.removeItem(self.QGitem)
                if self.line:
                    if self.line in self.items():
                        self.removeItem(self.line)

                    self.line.popPoint()
                self.polystatus = self.POLYREADY
                self.QGitem = None
                self.update()

    def refreshShapestoLabels(self, labelclass):
        labelpolygons = labelclass.polygons
        labelclass.polygons = []
        [labelclass.assignObject(shape) for shape in labelpolygons]
        return

    def initializeClasses(self, names, colors):
        for c, name in enumerate(names):
            if (c < len(self.labelclasses)):
                self.initializeClass(name, colors[c], labelclass=self.labelclasses[c])
                self.refreshShapestoLabels(labelclass=self.labelclasses[c])
            else:
                self.initializeClass(name, colors[c])

    def updateColor(self, color):
        self.shapeColor = color
        active_labelclass = self.labelclasses[self.labelmode]
        active_labelclass.fillColor = self.shapeColor
        self.refreshShapestoLabels(labelclass=active_labelclass)
        return

    def setLabelMode(self, classindex):
        labelclass = self.labelclasses[classindex]
        self.labelmode = classindex
        self.shapeColor = labelclass.fillColor
        return

    def initializeClass(self, name, color, labelclass=None):
        ind = None
        if labelclass is None:
            labelclass = LabelClass()
        else:
            ind = self.labelclasses.index(labelclass)
        labelclass.name = name
        labelclass.fillColor = color
        if ind is not None:
            self.labelclasses[ind] = labelclass
        else:
            self.labelclasses.append(labelclass)

    def triggerClosure(self):
        self.finalisepoly(premature=True)

    def mousePressEvent(self, event):
        '''Draw, move vertices/shapes, open properties window'''
        pos = event.scenePos()

        if (event.button() == Qt.RightButton) and not self.drawing():
            if self.selectedShape:
                all_labels = [None]
                if len(self.labelclasses) > 0:
                    all_labels = self.labelclasses
                propdialog = PropertiesWindow(shape=self.selectedShape, all_labels=all_labels, scene=self)
                propdialog.move(event.screenPos())
                propdialogexec = propdialog.exec()

                if propdialogexec:
                    if self.selectedShape.editable:
                        self.selectedShape.closed = False
                        self.QGitem = self.selectedShape
                        p = self.QGitem.points[-1]
                        self.line = Shape(point_size=self.point_size)
                        self.addItem(self.line)
                        self.line.setPos(p)
                        self.line.addPoint(p)
                        self.polystatus = self.POLYDRAWING
                        self.mode = self.DRAWING
                        shapeid = self.polys.index(self.selectedShape)
                        if self.selectedShape in self.polys:
                            self.polys.pop(shapeid)
                            self.objtypes.pop(shapeid)
                            self.QGitem.setZValue(len(self.polys) + 2)
                self.update()

        if self.drawing() & (event.button() == Qt.LeftButton):
            self.overrideCursor(CURSOR_DRAW)
            # update the tail of the pointing line
            if self.line and self.polygon_not_finished():
                self.line.prepareGeometryChange()
                self.line.points[0] = pos
                self.line.setPos(pos)
            # initialize a pointing line for a new polygon
            elif self.line == None or self.polygonfinished():
                self.line = Shape(point_size=self.point_size)
                self.addItem(self.line)
                self.line.setPos(pos)
                self.line.addPoint(pos)

            if self.QGitem:
                # attract the cursor to the start point of the polygon and close it
                self.QGitem.prepareGeometryChange()
                if len(self.QGitem.points) > 1 and self.closeEnough(pos, self.QGitem.points[0]):
                    pos = self.QGitem.points[0]
                    self.overrideCursor(CURSOR_POINT)
                    self.QGitem.highlightVertex(0)

                self.QGitem.addPoint(pos)
                if (self.QGitem.points[0] == pos):
                    self.finalisepoly()


            else:
                self.polystatus = self.POLYDRAWING
                self.QGitem = Shape(point_size=self.point_size)

                self.addItem(self.QGitem)
                self.QGitem.setPos(pos)
                self.QGitem.addPoint(pos)
                self.QGitem.setZValue(len(self.polys) + 1)
            self.update()
            event.accept()

        elif self.moving() & (event.button() == Qt.LeftButton):
            self.overrideCursor(CURSOR_GRAB)
            self.selectShapebyPoint(pos)
            self.prevPoint = pos
            event.accept()
            self.update()

        elif self.navigating():
            self.overrideCursor(CURSOR_GRAB)
            self.update()

    def mouseMoveEvent(self, event):
        '''Track the movement of the cursor and update selections/drawings'''
        pos = event.scenePos()

        if self.drawing():
            self.overrideCursor(CURSOR_DRAW)

            if self.QGitem:
                if len(self.QGitem.points) == 1:  # initialize the pointing line collapsed to a point
                    self.line.points = [self.QGitem.points[0], self.QGitem.points[0]]
                colorLine = self.lineColor
                if len(self.QGitem) > 1 and self.closeEnough(pos, self.QGitem[0]):
                    pos = self.QGitem[0]
                    colorLine = self.QGitem.line_color
                    self.overrideCursor(CURSOR_POINT)
                    self.QGitem.highlightVertex(0)

                if len(self.line.points) == 2:  # update the pointing line
                    self.line.points[1] = pos
                else:  # load the pointing line (if another shape was just created)
                    self.line.addPoint(pos)

                self.line.line_color = colorLine
                self.update()
            return

        # moving shapes/vertices
        if self.moving and Qt.LeftButton & event.buttons():
            self.overrideCursor(CURSOR_GRAB)
            if self.vertexSelected():
                self.selectedShape.prepareGeometryChange()
                if (self.selectedShape.objtype == 'Line') and (QApplication.keyboardModifiers() == Qt.ShiftModifier):
                    self.moveShape(self.selectedShape, pos)
                else:
                    self.moveVertex(pos)
                self.update()
            elif self.selectedShape and self.prevPoint:
                self.selectedShape.prepareGeometryChange()
                self.moveShape(self.selectedShape, pos)
                self.update()
            return

        # update selections/highlights based on cursor location

        # check if any vertex is epsilon close to the cursor position and find the corresponding shape
        id_point = [[i for i, y in enumerate(poly.points) if distance(pos - y) <= self.epsilon] for poly in self.polys]
        id_shape = [i for i, y in enumerate(id_point) if y != []]

        itemUnderMouse = self.itemAt(pos, QTransform())
        # self.clearShapeSelections()
        # if shape/vertex combination found, highlight vertex and shape
        if id_shape != []:
            self.selectedVertex = id_point[id_shape[0]][0]
            self.selectShape(self.items()[:-1][::-1][id_shape[0]])
            self.selectedShape.highlightVertex(self.selectedVertex)
            self.update()
            return
        elif itemUnderMouse in self.items()[:-1]:  # if the cursor is inside of a shape, highlight it
            self.selectedVertex = None
            self.selectShape(itemUnderMouse)
            self.selectedShape.hIndex = None
            self.update()
            return
        else:  # nothing found: no shape under the cursor, no vertices in vicinity, clear all
            self.selectedVertex = None
            self.update()
            return

        event.accept()

    def mouseReleaseEvent(self, event):
        if self.navigating or (event.button() == Qt.LeftButton and self.selectedShape):
            self.overrideCursor(CURSOR_DEFAULT)
            self.update()
        event.accept()

    def closeEnough(self, p1, p2):
        return distance(p1 - p2) < self.epsilon

    def finalisepoly(self, premature=False):
        if self.QGitem:
            if premature:
                if len(self.QGitem.points) == 1:
                    self.objtypes.append('Point')
                    self.QGitem.objtype = 'Point'
                else:
                    self.objtypes.append('Line')
                    self.QGitem.objtype = 'Line'
            else:
                self.objtypes.append('Polygon')
                self.QGitem.objtype = 'Polygon'
            if self.line:
                self.removeItem(self.line)
                self.line.popPoint()
            self.QGitem.editable = False
            self.polys.append(self.QGitem)
            if self.labelmode is not None:
                labelobject = self.labelclasses[self.labelmode]
                labelobject.assignObject(self.QGitem)
                self.QGitem.label = labelobject.name
            self.QGitem = None
            self.polystatus = self.POLYREADY
            self.update()

    def overrideCursor(self, cursor):
        self._cursor = cursor
        QApplication.setOverrideCursor(cursor)

    def copySelected(self):

        if self.selectedShape:
            shape = self.selectedShape
            c, o, s = [shape.closed, shape.objtype, shape.point_size]
            p = [point for point in shape.points]

            newshape = Shape()

            newshape.points, newshape.closed, newshape.objtype, newshape.point_size = p, c, o, s

            self.polys.append(newshape)
            self.objtypes.append(newshape.objtype)

            newshape.setZValue(len(self.polys) + 1)
            self.addItem(newshape)

            labelid = [label.name for label in self.labelclasses].index(shape.label)
            self.labelclasses[labelid].assignObject(newshape)

            print('Shape copied')
            self.clearShapeSelections()
            self.selectShape(newshape)
            self.update()
            return

    def deleteSelected(self):
        if self.selectedShape:
            shape = self.items()[:-1][::-1].index(self.selectedShape)
            if self.selectedShape in self.polys:
                self.polys.pop(shape)
                self.objtypes.pop(shape)
            self.removeItem(self.selectedShape)
            if self.line:
                if self.line in self.items():
                    self.removeItem(self.line)
                self.line.popPoint()
            labelind = self.findShapeInLabel(self.selectedShape)
            if len(labelind) > 0:
                label, shapeind = labelind[0]
                self.labelclasses[label].polygons.pop(shapeind)
            self.polystatus = self.POLYREADY
            self.selectedShape = None
            self.QGitem = None
            self.clearShapeSelections()
            print('Shape deleted')
            self.update()
            return

    def findShapeInLabel(self, shape):
        if len(self.labelclasses) > 0:
            labelpolys = [l.polygons for l in self.labelclasses]
            return ([(i, label.index(shape)) for i, label in enumerate(labelpolys) if shape in label])
        else:
            return 2 * [None]

    def selectShape(self, shape):
        shape.selected = True
        self.selectedShape = shape
        self.update()

    def selectShapebyPoint(self, point):
        """Select the first shape created which contains this point."""
        if self.vertexSelected():  # A vertex is marked for selection.
            self.selectedShape.highlightVertex(self.selectedVertex)
            return

        itemUnderMouse = self.itemAt(point, QTransform())

        if itemUnderMouse in self.items()[:-1]:
            self.selectShape(itemUnderMouse)
            return

    def clearShapeSelections(self):
        if self.selectedShape:
            self.selectedShape.highlightClear()
            self.selectedShape = None
            self.forceAllSelectionsClear()
            self.update()

    def forceAllSelectionsClear(self):
        for shape in self.items()[:-1]:
            shape.highlightClear()
        self.update()
        return

    def moveVertex(self, pos):
        self.selectedShape.prepareGeometryChange()
        self.selectedShape.moveBy(self.selectedVertex, pos - self.selectedShape[self.selectedVertex])

    def moveShape(self, shape, pos):
        delta = pos - self.prevPoint
        if delta:
            shape.prepareGeometryChange()
            shape.moveBy('all', delta)
            self.prevPoint = pos
            self.update()
            return True
        return False



import warnings
import numpy as np
import pandas as pd
import os
from PySide6.QtCore import Slot
from joblib import Parallel, delayed
from skimage import feature
from tqdm.autonotebook import tqdm
from ipywidgets import widgets
from ipywidgets import Layout, interact
from skimage.feature import peak_local_max

from piscat.Preproccessing import normalization
from piscat.InputOutput.cpu_configurations import CPUConfigurations
from piscat.Localization import data_handeling, frst
from piscat.Localization import gaussian_2D_fit
from piscat.Visualization import display_jupyter
from piscat.Preproccessing import filtering
from piscat.Localization import radial_symmetry_centering
from piscat.Visualization.display_jupyter import JupyterPSFs_localizationPreviewDisplay
from piscat.Localization.difference_of_gaussian import dog_preview


class PSFsExtraction:

    def __init__(self, video, flag_transform=False, flag_GUI=False, **kwargs):

        """
        This class employs a variety of PSF localization methods, including DoG/LoG/DoH/RS/RVT.

        It returns a list containing the following details:

        [[frame number, center y, center x, sigma], [frame number, center y, center x, sigma], ...]

        Parameters
        ----------
        video: NDArray
            Numpy 3D video array.

        flag_transform: bool
            In case it is defined as true, the input video is already transformed.
            So it does not need to run this task during localization.

        flag_GUI: bool
            While the GUI is calling this class, it is true.

        """
        super(PSFsExtraction, self).__init__()
        self.cpu = CPUConfigurations()
        self.kwargs = kwargs
        self.counter = 0

        self.flag_GUI = flag_GUI

        self.video = video
        self.flag_transform = flag_transform
        self.norm_vid = None

        self.psf_dog = None
        self.psf_doh = None
        self.psf_log = None
        self.psf_hog = None
        self.psf_frst = None

        self.psf_hog_1D_feature = None

        self.min_sigma = None
        self.max_sigma = None
        self.sigma_ratio = None
        self.threshold = None
        self.overlap = None
        self.radii = None
        self.alpha = None
        self.beta = None
        self.stdFactor = None
        self.mode = None
        self.function = None

        self.df_PSF = None

    @Slot()
    def run(self):
        result = self.psf_detection(**self.kwargs)
        self.signals.result.emit(result)

    def dog(self, image):
        """
        PSF localization using DoG.

        Parameters
        ----------
        image: NDArray
            image is an input numpy array.

        Returns
        -------
        tmp: list
            [y, x, sigma]
        """

        return feature.blob_dog(image, min_sigma=self.min_sigma, max_sigma=self.max_sigma, sigma_ratio=self.sigma_ratio,
                                threshold=self.threshold, overlap=self.overlap, exclude_border=True)

    def doh(self, image):
        """
        PSF localization using DoH.

        Parameters
        ----------
        image: NDArray
            image is an input numpy array.

        Returns
        -------
        tmp: list
            [y, x, sigma]
        """

        tmp = feature.blob_doh(image, min_sigma=self.min_sigma, max_sigma=self.max_sigma,
                               num_sigma=int(self.sigma_ratio),
                               threshold=self.threshold, overlap=self.overlap)
        return tmp

    def log(self, image):
        """
        PSF localization using LoG.

        Parameters
        ----------
        image: NDArray
            image is an input numpy array.

        Returns
        -------
        @returns: list
            [y, x, sigma]
        """

        return feature.blob_log(image, min_sigma=self.min_sigma, max_sigma=self.max_sigma,
                                num_sigma=int(self.sigma_ratio),
                                threshold=self.threshold, overlap=self.overlap, exclude_border=True)

    def frst(self, image):
        """
        PSF localization using frst.

        Parameters
        ----------
        image: NDArray
            image is an input numpy array.

        Returns
        -------
        tmp: list
            [y, x, sigma]

        References
        ----------
            [1] Loy, G., & Zelinsky, A. (2002). A fast radial symmetry transform for detecting points of interest. Computer Vision, ECCV 2002.
        """

        normaliz_image = normalization.Normalization(image).normalized_image()

        tmp = frst.blob_frst(normaliz_image, min_radial=self.min_radial, max_radial=self.max_radial,
                             radial_step=self.radial_step, threshold=self.threshold, alpha=self.alpha, beta=self.beta,
                             stdFactor=self.stdFactor, mode=self.mode)
        return tmp

    def _rvt(self, image):
        """
        PSF localization using RVT.

        Parameters
        ----------
        image: NDArray
            image is an input numpy array.

        Returns
        -------
        tmp: list
            [y, x, sigma]
        """
        if self.flag_transform:
            tr_img = image
        else:
            rvt_ = filtering.RadialVarianceTransform()
            tr_img = rvt_.rvt(img=image, rmin=self.min_radial, rmax=self.max_radial, kind=self.rvt_kind, highpass_size=self.highpass_size,
                        upsample=self.upsample, rweights=self.rweights, coarse_factor=self.coarse_factor, coarse_mode=self.coarse_mode,
                        pad_mode=self.pad_mode)

        local_maxima = peak_local_max(
            tr_img,
            threshold_abs=self.threshold,
            footprint=np.ones((3,) * (image.ndim)),
            threshold_rel=0.0,
            exclude_border=True,
        )

        # Catch no peaks
        if local_maxima.size == 0:
            return np.empty((0, 3))

        sigmas = (self.min_radial/np.sqrt(2)) * np.ones((local_maxima.shape[0], 1))
        tmp = np.concatenate((local_maxima, sigmas), axis=1)

        return tmp

    def fit_Gaussian2D_wrapper(self, PSF_List, scale=5, internal_parallel_flag=False):
        """
        PSF localization using fit_Gaussian2D.

        Parameters
        ----------
        PSF_List: pandas dataframe
            The data frame contains PSFs locations( x, y, frame, sigma)

        scale: int
            The ROI around PSFs is defined using this scale, which is based on their sigmas.

        internal_parallel_flag: bool
            Internal flag for activating parallel computation. Default is True!

        Returns
        -------
        df: pandas dataframe
            The data frame contains PSFs locations ( 'y', 'x', 'frame', 'center_intensity', 'sigma', 'Sigma_ratio') and fitting information.
            fit_params is a list include ('Fit_Amplitude', 'Fit_X-Center', 'Fit_Y-Center', 'Fit_X-Sigma', 'Fit_Y-Sigma',
            'Fit_Bias', 'Fit_errors_Amplitude', 'Fit_errors_X-Center', 'Fit_errors_Y-Center', 'Fit_errors_X-Sigma', 'Fit_errors_Y-Sigma', 'Fit_errors_Bias'].
        """

        if type(PSF_List) is list:
            df_PSF = data_handeling.list2dataframe(feature_position=PSF_List, video=self.video)
        elif type(PSF_List) is pd.core.frame.DataFrame:
            df_PSF = PSF_List
        else:
            raise ValueError('PSF_List does not have correct bin_type')

        self.df2numpy = df_PSF.to_numpy()

        if self.cpu.parallel_active and internal_parallel_flag:
            print('\n---Fitting 2D gaussian with parallel loop---')
            list_df = Parallel(n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose)(delayed(
                self.fit_2D_gussian_kernel)(i_, scale) for i_ in tqdm(range(self.df2numpy.shape[0])))

        else:
            print('\n---Fitting 2D gaussian without parallel loop---')
            list_df = []
            for i_ in tqdm(range(self.df2numpy.shape[0])):
                tmp = self.fit_2D_gussian_kernel(i_, scale)
                list_df.append(tmp)

        df2numpy = np.asarray(list_df)

        if df2numpy.shape[0] != 0:
            df = pd.DataFrame(data=df2numpy, columns=['y', 'x', 'frame', 'center_intensity', 'sigma', 'Sigma_ratio',
                                                       'Fit_Amplitude', 'Fit_X-Center', 'Fit_Y-Center', 'Fit_X-Sigma',
                                                       'Fit_Y-Sigma', 'Fit_Bias', 'Fit_errors_Amplitude',
                                                       'Fit_errors_X-Center', 'Fit_errors_Y-Center', 'Fit_errors_X-Sigma',
                                                       'Fit_errors_Y-Sigma', 'Fit_errors_Bias'])
        else:
            df = None
        return df

    def fit_2D_gussian_kernel(self, i_, scale=5, flag_init=False, image=None, start_sigma=[None, None], display_flag=False):

        if flag_init:
            fit_params_ = gaussian_2D_fit.fit_2D_Gaussian_varAmp(image, sigma_x=start_sigma[0],
                                                                 sigma_y=start_sigma[1],
                                                                 display_flag=display_flag)
            return fit_params_
        else:

            sigma_0 = self.df2numpy[i_, 4]
            cen_int = self.df2numpy[i_, 3]
            frame_num = self.df2numpy[i_, 2]
            p_x = self.df2numpy[i_, 1]
            p_y = self.df2numpy[i_, 0]

            window_size = scale * np.sqrt(2) * sigma_0
            start_sigma = sigma_0

            img_size_x = self.video.shape[2]
            img_size_y = self.video.shape[1]

            start_y = np.max([0, p_y - window_size])
            start_x = np.max([0, p_x - window_size])

            end_y = np.min([img_size_y-1, p_y + window_size])
            end_x = np.min([img_size_x-1, p_x + window_size])

            w_s_x_1 = round(abs(p_x - start_x))
            w_s_y_1 = round(abs(p_y - start_y))

            w_s_x_2 = round(abs(p_y - end_y))
            w_s_y_2 = round(abs(p_x - end_x))

            w_s = np.min([w_s_x_1, w_s_y_1, w_s_x_2, w_s_y_2])

            start_y = p_y - w_s
            start_x = p_x - w_s

            end_y = p_y + w_s
            end_x = p_x + w_s

            if start_y >= 0 and start_x >= 0 and end_y <= img_size_y and end_x <= img_size_x:

                window_frame = self.video[int(frame_num), int(start_y):int(end_y),
                               int(start_x):int(end_x)]
            else:
                raise ValueError('Cropping size is not accurate!')

            fit_params = gaussian_2D_fit.fit_2D_Gaussian_varAmp(window_frame, sigma_x=start_sigma,
                                                                sigma_y=start_sigma,
                                                                display_flag=display_flag)

            params = [p_y, p_x, frame_num, cen_int, sigma_0]
            if fit_params[0] is not None:
                params.append(fit_params[0])
            else:
                params.append(np.nan)

            if fit_params[1] is not None:
                params.append(fit_params[1][0])
                params.append(fit_params[1][1] + p_x - w_s)
                params.append(fit_params[1][2] + p_y - w_s)
                params.append(fit_params[1][3])
                params.append(fit_params[1][4])
                params.append(fit_params[1][5])
            else:
                params.append(np.nan)
                params.append(np.nan)
                params.append(np.nan)
                params.append(np.nan)
                params.append(np.nan)
                params.append(np.nan)

            if fit_params[2] is not None:
                params.append(fit_params[2][0])
                params.append(fit_params[2][1])
                params.append(fit_params[2][2])
                params.append(fit_params[2][3])
                params.append(fit_params[2][4])
                params.append(fit_params[2][5])
            else:
                params.append(np.nan)
                params.append(np.nan)
                params.append(np.nan)
                params.append(np.nan)
                params.append(np.nan)
                params.append(np.nan)
            return params

    def frst_one_PSF(self, image):
        """
        The lateral position of PSFs with subpixel resolution is returned by this function.
        Because this function only works when there is only one PSF in the field of view, it
        is typically used after coarse localization to extract fine localization for each PSF.

        Parameters
        ----------
        image: NDArray
            image is an input numpy array.

        Returns
        -------
        @returns: list
            [y, x, sigma]

        References
        ----------
            [1] Parthasarathy, R. Rapid, accurate particle tracking by calculation of radial symmetry centers.
            Nat Methods 9, 724–726 (2012). https://doi.org/10.1038/nmeth.2071
        """
        if image.shape[0] == image.shape[1]:
            xc, yc, sigma = radial_symmetry_centering.RadialCenter().radialcenter(Image=image)

        elif image.shape[0] > image.shape[1]:
            tmp = np.median(image) + np.zeros((image.shape[0], image.shape[0]))
            tmp[:image.shape[0], :image.shape[1]] = image
            offset = image.shape[0] - image.shape[1]
            xc, yc, sigma = radial_symmetry_centering.RadialCenter().radialcenter(Image=tmp)
            xc = xc - offset

        elif image.shape[0] < image.shape[1]:
            tmp = np.median(image) + np.zeros((image.shape[1], image.shape[1]))
            tmp[:image.shape[0], :image.shape[1]] = image
            offset = image.shape[1] - image.shape[0]
            xc, yc, sigma = radial_symmetry_centering.RadialCenter().radialcenter(Image=tmp)
            yc = yc - offset

        return [yc, xc, sigma]

    def improve_localization_with_frst(self, df_PSFs, scale, flag_preview=False):
        """
        It extracts subpixels localization based on initial pixel localization using ``frst_one_PSF`` methods for all detected PSFs.

        Parameters
        ----------
        df_PSFs: pandas dataframe
            The data frame contains PSFs locations( x, y, frame, sigma)

        scale: int
            The ROI around PSFs is defined using this scale, which is based on their sigmas.

        flag_preview: bool
            When the GUI calls these functions, this flag is set as True.

        Returns
        -------
        sub_pixel_localization: pandas dataframe
            The data frame contains subpixels PSFs locations( x, y, frame, sigma)
        """
        sigma = df_PSFs['sigma'].tolist()
        psf_position_x = df_PSFs['x'].tolist()
        psf_position_y = df_PSFs['y'].tolist()
        psf_position_frame = df_PSFs['frame'].tolist()

        if self.cpu.parallel_active and flag_preview is not True:
            print("\n---start improve_localization with parallel loop---")
            try:
                result0 = Parallel(n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose)(
                    delayed(self.frst_wrapper)(p_, scale) for p_ in tqdm(zip(psf_position_frame, psf_position_x, psf_position_y, sigma)))

            except:
                warnings.warn("switch backend to threading", DeprecationWarning)
                result0 = Parallel(n_jobs=self.cpu.n_jobs, backend='threading', verbose=self.cpu.verbose)(
                    delayed(self.frst_wrapper)(p_, scale) for p_ in tqdm(zip(psf_position_frame, psf_position_x, psf_position_y, sigma)))

        else:
            print("\n---start improve_localization without parallel loop---")

            result0 = [self.frst_wrapper(p_, scale) for p_ in tqdm(zip(psf_position_frame, psf_position_x, psf_position_y, sigma))]

        tmp2 = [tmp for tmp in result0 if isinstance(tmp, np.ndarray)]  # remove non values
        sub_pixel_localization = data_handeling.list2dataframe(feature_position=tmp2, video=self.video)

        return sub_pixel_localization

    def frst_wrapper(self, p_, scale=4):
        subPixel = None

        f_ = int(p_[0])
        x_ = int(p_[1])
        y_ = int(p_[2])
        sigma_ = p_[3]
        win_size = int(scale * sigma_)

        tmp_x = [0, (x_ - win_size)]
        x_start = np.max(tmp_x)

        tmp_x = [(x_ + win_size), self.video.shape[1]]
        x_end = np.min(tmp_x)

        tmp_y = [0, (y_ - win_size)]
        y_start = np.max(tmp_y)

        tmp_y = [(y_ + win_size), self.video.shape[2]]
        y_end = np.min(tmp_y)

        particle_windows = self.video[f_, y_start:y_end, x_start:x_end]
        if particle_windows.shape[0] == particle_windows.shape[1]:
            new_position = self.frst_one_PSF(particle_windows)

            if (new_position[0] + y_ - win_size) > self.video.shape[1]:
                l_y = None
            else:
                l_y = new_position[0] + y_ - win_size

            if (new_position[1] + x_ - win_size) > self.video.shape[2]:
                l_x = None
            else:
                l_x = new_position[1] + x_ - win_size

            if l_x is not None and l_y is not None:
                tmp = np.expand_dims(np.asarray([f_, l_y, l_x, new_position[2]]), axis=0)

                if subPixel is None:
                    subPixel = tmp
                else:
                    subPixel = np.concatenate((subPixel, tmp), axis=0)

            return subPixel

    def psf_detection(self, function, min_sigma=1, max_sigma=2, sigma_ratio=1.1, threshold=0, overlap=0,
                      min_radial=1, max_radial=2, radial_step=0.1, alpha=2, beta=1, stdFactor=1,
                      rvt_kind="basic",  highpass_size=None, upsample=1, rweights=None, coarse_factor=1, coarse_mode="add",
                      pad_mode="constant", mode='BOTH', flag_GUI_=False):
        """
        This function is a wrapper for calling various PSF localization methods.

        Parameters
        ----------
        function: str
            PSF localization algorithm which should be selected  from : (``'dog'``, ``'log'``, ``'doh'``, ``'frst'``,
            ``'frst_one_psf'``, ``'RVT'``)

        mode: str
            Defines which PSFs will be detected (``'BRIGHT'``, ``'DARK'``, or ``'BOTH'``).

        flag_GUI_: bool
            Only is used when GUI calls this function.

        optional_1:
            These parameters are used when ``'dog'``, ``'log'``, ``'doh'`` are defined as function.

            * `min_sigma`: float, list of floats
                The is the minimum standard deviation for the kernel. The lower the value, the smaller blobs can be detected.
                The standard deviations of the filter are given for each axis in sequence or with a single number which is considered for both axis.

            * `max_sigma`: float, list of floats
                The is the maximum standard deviation for the kernel. The higher the value, the bigger blobs can be detected.
                The standard deviations of the filter are given for each axis in sequence or with a single number
                which is considered for both axis.

            * `sigma_ratio`: float
                * The ratio between the standard deviation of Kernels which is used for computing the DoG and LoG.
                * The number of intermediate values of standard deviations between min_sigma and max_sigma for computing the DoH.

            * `threshold_min`: float
                The absolute lower bound for scale space maxima. Local maxima smaller than thresh are ignored. Reduce this
                to detect blobs with less intensities.

            * `overlap`: float
                A value between 0 and 1. If the area of two blobs are overlapping by a fraction greater than threshold_min, smaller blobs are eliminated.

        optional_2:
            These parameters are used when ``'frst'`` is defined as function.

            * `min_radial`: int
                integer value for radius size in pixels (n in the original paper); also is used as gaussian kernel size

            * `max_radial`: int
                integer value for radius size in pixels (n in the original paper); also is used as gaussian kernel size

            * `alpha`: str
                Strictness of symmetry transform (higher=more strict; 2 is good place to start)

            * `beta`: float
                gradient threshold_min parameter, float in range [0,1]

            * `stdFactor`:
                Standard deviation factor for gaussian kernel

        optional_3:
            These parameters are used when ``"RVT"`` is defined as function.

            * `min_radial`:
                minimal radius (inclusive)

            * `max_radial`:
                maximal radius (inclusive)

            * `rvt_kind`:
                either ``"basic"`` (only VoM), or ``"normalized"`` (VoM/MoV);
                normalized version increases subpixel bias, but it works better at lower SNR

            * `highpass_size`:
                size of the high-pass filter; ``None`` means no filter (effectively, infinite size)

            * `upsample`: int
                integer image upsampling factor;
                `rmin` and `rmax` are adjusted automatically (i.e., they refer to the non-upsampled image);
                if ``upsample>1``, the resulting image size is multiplied by ``upsample``

            * `rweights`:
                relative weights of different radial kernels; must be a 1D array of the length ``(rmax-rmin+1)//coarse_factor``
                coarse_factor: the reduction factor for the number ring kernels; can be used to speed up calculations at the expense of precision
                coarse_mode: the reduction method; can be ``"add"`` (add ``coarse_factor`` rings in a row to get a thicker ring, which works better for smooth features),
                or ``"skip"`` (only keep on in ``coarse_factor`` rings, which works better for very fine features)

            * `coarse_factor`:
                the reduction factor for the number ring kernels; can be used to speed up calculations at the expense of precision

            * `coarse_mode`:
                the reduction method; can be ``"add"`` (add ``coarse_factor`` rings in a row to get a thicker ring, which works better for smooth features),
                or ``"skip"`` (only keep on in ``coarse_factor`` rings, which works better for very fine features)

            * `pad_mode`:
                edge padding mode for convolutions; can be either one of modes accepted by ``np.pad`` (such as ``"constant"``, ``"reflect"``, or ``"edge"``),
                or ``"fast"``, which means faster no-padding (a combination of ``"wrap"`` and ``"constant"`` padding depending on the image size);
                ``"fast"`` mode works faster for smaller images and larger ``rmax``, but the border pixels (within ``rmax`` from the edge) are less reliable;
                note that the image mean is subtracted before processing, so ``pad_mode="constant"`` (default) is equivalent to padding with a constant value equal to the image mean

        Returns
        -------
        df_PSF: pandas dataframe
            The dataframe for PSFs that contains the ['x', 'y', 'frame number', 'sigma'] for each PSF.
        """

        self.min_sigma = min_sigma
        self.max_sigma = max_sigma
        self.sigma_ratio = sigma_ratio
        self.threshold = threshold
        self.overlap = overlap
        self.min_radial = min_radial
        self.max_radial = max_radial
        self.radial_step = radial_step
        self.alpha = alpha
        self.beta = beta
        self.stdFactor = stdFactor
        self.mode = mode

        self.rvt_kind = rvt_kind
        self.highpass_size = highpass_size
        self.upsample = upsample
        self.rweights = rweights
        self.coarse_factor = coarse_factor
        self.coarse_mode = coarse_mode
        self.pad_mode = pad_mode

        if self.video.ndim != 3 and self.video.ndim == 2:
            self.video = np.expand_dims(self.video, axis=0)

        if self.cpu.parallel_active:
            self.function = function

            print("\n---start PSF detection with parallel loop---")
            if flag_GUI_ is False:
                try:
                    result0 = Parallel(n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose)(
                        delayed(self.psf_detection_kernel)(x) for x in tqdm(range(self.video.shape[0])))
                    result = [x for x in result0 if x is not None]
                except:
                    warnings.warn("switch backend to threading", DeprecationWarning)

                    result0 = Parallel(n_jobs=self.cpu.n_jobs, backend='threading', verbose=self.cpu.verbose)(
                        delayed(self.psf_detection_kernel)(x) for x in range(self.video.shape[0]))
                    result = [x for x in result0 if x is not None]
            else:
                result0 = Parallel(n_jobs=self.cpu.n_jobs, backend='threading', verbose=self.cpu.verbose)(
                    delayed(self.psf_detection_kernel)(x) for x in tqdm(range(self.video.shape[0])))
                result = [x for x in result0 if x is not None]

        else:
            print("\n---start PSF detection without parallel loop---")
            self.function = function
            result0 = [self.psf_detection_kernel(x) for x in tqdm(range(self.video.shape[0]))]
            result = [x for x in result0 if x is not None]

        df_PSF = data_handeling.list2dataframe(feature_position=result, video=self.video)

        return df_PSF

    def psf_detection_preview(self,  function, min_sigma=1, max_sigma=2, sigma_ratio=1.1, threshold=0, overlap=0,
                                min_radial=1, max_radial=2, radial_step=0.1, alpha=2, beta=1, stdFactor=1,
                                rvt_kind="basic",  highpass_size=None, upsample=1, rweights=None, coarse_factor=1, coarse_mode="add",
                                pad_mode="constant", mode='BOTH', frame_number=0, median_filter_flag=False, color='gray',
                                imgSizex=5, imgSizey=5, IntSlider_width='500px', title=''):

        """
        This function is a preview wrapper for calling various PSF localization methods.

        Parameters
        ----------
        function: str
            PSF localization algorithm which should be selected  from : (``'dog'``, ``'log'``, ``'doh'``, ``'frst'``, ``'frst_one_psf``')

        mode: str
            Defines which PSFs will be detected (``'BRIGHT'``, ``'DARK'``, or ``'BOTH'``).

        frame_number: int
            Selecting frame number that PSFs detection should apply on it.

        optional_1:
            These parameters are used when ``'dog'``, ``'log'``, ``'doh'`` are defined as function.

            * `min_sigma`: float, list of floats
                The is the minimum standard deviation for the kernel. The lower the value, the smaller blobs can be detected.
                The standard deviations of the filter are given for each axis in sequence or with a single number which is considered for both axis.

            * `max_sigma`: float, list of floats
                The is the maximum standard deviation for the kernel. The higher the value, the bigger blobs can be detected.
                The standard deviations of the filter are given for each axis in sequence or with a single number
                which is considered for both axis.

            * `sigma_ratio`: float
                * The ratio between the standard deviation of Kernels which is used for computing the DoG and LoG.
                * The number of intermediate values of standard deviations between min_sigma and max_sigma for computing the DoH.

            * `threshold_min`: float
                The absolute lower bound for scale space maxima. Local maxima smaller than thresh are ignored. Reduce this
                to detect blobs with less intensities.

            * `overlap`: float
                A value between 0 and 1. If the area of two blobs are overlapping by a fraction greater than threshold_min, smaller blobs are eliminated.

        optional_2:
            These parameters are used when ``'frst'`` is defined as function.

            * `min_radial`: int
                integer value for radius size in pixels (n in the original paper); also is used as gaussian kernel size

            * `max_radial`: int
                integer value for radius size in pixels (n in the original paper); also is used as gaussian kernel size

            * `alpha`: str
                Strictness of symmetry transform (higher=more strict; 2 is good place to start)

            * `beta`: float
                gradient threshold_min parameter, float in range [0,1]

            * `stdFactor`:
                Standard deviation factor for gaussian kernel

        optional_3:
            These parameters are used when ``'RVT'`` is defined as function.

            * `min_radial`:
                minimal radius (inclusive)

            * `max_radial`:
                maximal radius (inclusive)

            * `rvt_kind`:
                either ``"basic"`` (only VoM), or ``"normalized"`` (VoM/MoV);
                normalized version increases subpixel bias, but it works better at lower SNR

            * `highpass_size`:
                size of the high-pass filter; ``None`` means no filter (effectively, infinite size)

            * `upsample`: int
                integer image upsampling factor;
                `rmin` and `rmax` are adjusted automatically (i.e., they refer to the non-upsampled image);
                if ``upsample>1``, the resulting image size is multiplied by ``upsample``

            * `rweights`:
                relative weights of different radial kernels; must be a 1D array of the length ``(rmax-rmin+1)//coarse_factor``
                coarse_factor: the reduction factor for the number ring kernels; can be used to speed up calculations at the expense of precision
                coarse_mode: the reduction method; can be ``"add"`` (add ``coarse_factor`` rings in a row to get a thicker ring, which works better for smooth features),
                or ``"skip"`` (only keep on in ``coarse_factor`` rings, which works better for very fine features)

            * `coarse_factor`:
                the reduction factor for the number ring kernels; can be used to speed up calculations at the expense of precision

            * `coarse_mode`:
                the reduction method; can be ``"add"`` (add ``coarse_factor`` rings in a row to get a thicker ring, which works better for smooth features),
                or ``"skip"`` (only keep on in ``coarse_factor`` rings, which works better for very fine features)

            * `pad_mode`:
                edge padding mode for convolutions; can be either one of modes accepted by ``np.pad`` (such as ``"constant"``, ``"reflect"``, or ``"edge"``),
                or ``"fast"``, which means faster no-padding (a combination of ``"wrap"`` and ``"constant"`` padding depending on the image size);
                ``"fast"`` mode works faster for smaller images and larger ``rmax``, but the border pixels (within ``rmax`` from the edge) are less reliable;
                note that the image mean is subtracted before processing, so ``pad_mode="constant"`` (default) is equivalent to padding with a constant value equal to the image mean

        Returns
        -------
        df_PSF: pandas dataframe
            The dataframe for PSFs that contains the ['x', 'y', 'frame number', 'sigma'] for each PSF
        """

        self.min_sigma = min_sigma
        self.max_sigma = max_sigma
        self.sigma_ratio = sigma_ratio
        self.overlap = overlap
        self.min_radial = min_radial
        self.max_radial = max_radial
        self.radial_step = radial_step
        self.alpha = alpha
        self.beta = beta
        self.stdFactor = stdFactor
        self.mode = mode
        self.function = function
        self.frame_number = frame_number
        self.title = title

        self.rvt_kind = rvt_kind
        self.highpass_size = highpass_size
        self.upsample = upsample
        self.rweights = rweights
        self.coarse_factor = coarse_factor
        self.coarse_mode = coarse_mode
        self.pad_mode = pad_mode

        self.median_filter_flag = median_filter_flag
        self.color = color
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.IntSlider_width = IntSlider_width

        if self.video.ndim != 3 and self.video.ndim == 2:
            self.video = np.expand_dims(self.video, axis=0)

        if "JPY_PARENT_PID" in os.environ:

            display_ = display_jupyter.JupyterPSFs_localizationPreviewDisplay(video=self.video, df_PSFs=None,
                                                              frame_num=self.frame_number, title=self.title,
                                                               median_filter_flag=self.median_filter_flag,
                                                               color=self.color, imgSizex=self.imgSizex,
                                                               imgSizey=self.imgSizey,
                                                               IntSlider_width=self.IntSlider_width)
            def _preview(threshold):
                self.threshold = threshold
                df_PSF = self.psf_preview_kernel()

                display_.set_df(df_PSF)
                display_.display_run()

            selected_frame = self.video[self.frame_number, ...]
            min_range_, max_range_ = dog_preview(images=selected_frame, min_sigma=self.min_sigma,
                                                 max_sigma=self.max_sigma, sigma_ratio=self.sigma_ratio)

            max_range_ = np.max(np.asarray([max_range_, threshold]))

            max_range_0 = 1.5 * max_range_
            sci_num = lambda x: "{:.2e}".format(x)
            tmp = sci_num(max_range_).split('e')
            power = int(tmp[-1])
            if power >= 0:
                step = 0.1
            elif power <= -3:
                step = 1e-5
            else:
                p_ = power - 2
                step = 10**p_

            interact(_preview,
                     threshold=widgets.FloatSlider(value=threshold, min=0, max=max_range_0, step=step,
                                                   continuous_update=False, readout=True, readout_format='.7f',
                                                    description='Threshold', layout=Layout(width=IntSlider_width)))
        else:
            self.threshold = threshold
            df_PSF = self.psf_preview_kernel()
            return df_PSF

    def psf_preview_kernel(self):

        if type(self.frame_number) == list:
            result = [self.psf_detection_kernel(f_) for f_ in self.frame_number]
        else:
            result = self.psf_detection_kernel(self.frame_number)

        df_PSF = data_handeling.list2dataframe(feature_position=result, video=self.video)

        return df_PSF

    def psf_detection_kernel(self, i_):
        i_ = int(i_)

        if len(self.video.shape) == 3 and self.video.shape[0] > 0:

            if self.function == 'dog':

                if self.mode == 'BOTH':
                    positive_psf = self.dog(self.video[i_, :, :])
                    negative_psf = self.dog(-1 * self.video[i_, :, :])
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)
                elif self.mode == 'Bright':
                    positive_psf = self.dog(self.video[i_, :, :])
                    negative_psf = []
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)
                elif self.mode == 'Dark':
                    positive_psf = []
                    negative_psf = self.dog(-1 * self.video[i_, :, :])
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)

            elif self.function == 'doh':

                if self.mode == 'BOTH':
                    positive_psf = self.doh(self.video[i_, :, :])
                    negative_psf = []
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)

            elif self.function == 'log':

                if self.mode == 'BOTH':
                    positive_psf = self.log(self.video[i_, :, :])
                    negative_psf = self.log(-1 * self.video[i_, :, :])
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)
                elif self.mode == 'Bright':
                    positive_psf = self.log(self.video[i_, :, :])
                    negative_psf = []
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)
                elif self.mode == 'Dark':
                    positive_psf = []
                    negative_psf = self.log(-1 * self.video[i_, :, :])
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)

            elif self.function == 'frst':
                b_psf = self.frst(self.video[i_, :, :])
                temp2 = self.concatenateBrightDark(b_psf, [], i_)

            elif self.function == 'frst_one_psf':
                b_psf = self.frst_one_PSF(self.video[i_, :, :])
                temp2 = self.concatenateBrightDark(b_psf, [], i_)
                temp2 = np.expand_dims(temp2, axis=0)

            elif self.function == 'RVT':
                b_psf = self._rvt(self.video[i_, :, :])
                temp2 = self.concatenateBrightDark(b_psf, [], i_)

        else:

            if self.function == 'dog':

                if self.mode == 'BOTH':
                    positive_psf = self.dog(self.video)
                    negative_psf = self.dog(-1 * self.video)
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)
                elif self.mode == 'Bright':
                    positive_psf = self.dog(self.video)
                    negative_psf = []
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)
                elif self.mode == 'Dark':
                    positive_psf = []
                    negative_psf = self.dog(-1 * self.video)
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)

            elif self.function == 'doh':

                if self.mode == 'BOTH':
                    positive_psf = self.doh(self.video)
                    negative_psf = []
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)

            elif self.function == 'log':

                if self.mode == 'BOTH':
                    positive_psf = self.log(self.video)
                    negative_psf = self.log(-1 * self.video)
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)
                elif self.mode == 'Bright':
                    positive_psf = self.log(self.video)
                    negative_psf = []
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)
                elif self.mode == 'Dark':
                    positive_psf = []
                    negative_psf = self.log(-1 * self.video)
                    temp2 = self.concatenateBrightDark(positive_psf, negative_psf, i_)

            elif self.function == 'frst':
                b_psf = self.frst(self.video)
                temp2 = self.concatenateBrightDark(b_psf, [], i_)

            elif self.function == 'frst_one_psf':
                b_psf = self.frst_one_PSF(self.video)
                temp2 = self.concatenateBrightDark(b_psf, [], i_)
                temp2 = np.expand_dims(temp2, axis=0)

            elif self.function == 'RVT':
                b_psf = self._rvt(self.video)
                temp2 = self.concatenateBrightDark(b_psf, [], i_)
                temp2 = np.expand_dims(temp2, axis=0)

        return temp2

    def concatenateBrightDark(self, bright_psf, dark_psf, i_):
        if len(bright_psf) != 0 and len(dark_psf) != 0:
            psf = np.unique(np.concatenate((bright_psf, dark_psf)), axis=0)

            frame_num = i_ * np.ones((psf.shape[0], 1), dtype=int)
            temp2 = np.concatenate((frame_num, psf), axis=1)
        elif len(bright_psf) != 0 and len(dark_psf) == 0:
            psf = np.asarray(bright_psf)
            frame_num = i_ * np.ones((psf.shape[0], 1), dtype=int)

            if len(psf.shape) != len(frame_num.shape) and len(psf.shape) == 1:
                temp2 = np.concatenate(([i_], psf), axis=0)
            else:
                temp2 = np.concatenate((frame_num, psf), axis=1)

        elif len(bright_psf) == 0 and len(dark_psf) != 0:
            psf = np.asarray(dark_psf)

            frame_num = i_ * np.ones((psf.shape[0], 1), dtype=int)
            temp2 = np.concatenate((frame_num, psf), axis=1)
        else:
            temp2 = None
        return temp2







from __future__ import print_function

import matplotlib.pyplot as plt
import numpy as np
import os
from matplotlib import pyplot as plt
import matplotlib.cm as cm
from matplotlib.animation import FuncAnimation


def plot3(X, Y, Z, scale='(um)', title=''):
    """
    3D localization plotting

    Parameters
    ----------
    X: list
        X-position of PSFs.

    Y: list
        Y-position of PSFs.

    Z: list
        Z-position of PSFs.

    scale: str
        Measurement scale. e.g ('nm')

    title: str
        The title of the figure.

    """
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    ax.scatter(X, Y, Z)

    ax.set_xlabel('X' + scale)
    ax.set_ylabel('Y' + scale)
    ax.set_zlabel('Z' + scale)

    plt.title(title)
    plt.show()


def plot2df(df_PSFs, pixel_size=1, scale='(um)', title='', flag_label=False):
    """
    2D localization plotting with color code for each particle.

    Parameters
    ----------
    df_PSFs: panda data frame
        Data frame contains PSFs localization and ID.

    pixel_size:
        The camera's pixel size is used to scale the results of the localization.

    scale: str
        Measurement scale. e.g ('nm')

    title: str
        The title of the figure.

    flag_label: bool


    """
    groups = df_PSFs.groupby('particle')

    fig, ax = plt.subplots()
    for name, group in groups:
        ax.plot(group.x*pixel_size, group.y*pixel_size, marker='.', linestyle='', ms=3, label=name)
    if flag_label:
        ax.legend()
    ax.set_xlabel(scale)
    ax.set_ylabel(scale)
    plt.title(title)
    plt.show()


def plot_bright_dark_psf(df_bright, df_dark, unit='nm'):
    """
    Plot heatmap of particle localization.

    Parameters
    ----------
    df_bright: pandas dataframe
        Bright PSF positions are stored in the data frame.

    df_dark: pandas dataframe
        Dark PSF positions are stored in the data frame.

    unit: str
        The axis unit.
    """
    particle_ID_bright = df_bright['particle'].tolist()
    particle_x_bright = df_bright['x'].tolist()
    particle_y_bright = df_bright['y'].tolist()
    particle_size_bright = df_bright['bubble_size'].tolist()

    particle_ID_dark = df_dark['particle'].tolist()
    particle_x_dark = df_dark['x'].tolist()
    particle_y_dark = df_dark['y'].tolist()
    particle_size_dark = df_dark['bubble_size'].tolist()

    particle_ID_bright_ = [str(l_b) for l_b in particle_ID_bright]
    particle_ID_dark_ = [str(l_b) for l_b in particle_ID_dark]

    particle_ID_all = particle_ID_bright_ + particle_ID_dark_
    particle_x_all = particle_x_bright + particle_x_dark
    particle_y_all = particle_y_bright + particle_y_dark
    particle_size_all = particle_size_bright + particle_size_dark

    fig = plt.figure()
    ax1 = plt.subplot(1, 1, 1)
    line1, = plt.plot(particle_x_bright, particle_y_bright, 'r.', markersize=15)

    ax2 = ax1.twinx()
    line2, = ax2.plot(particle_x_dark, particle_y_dark, 'b.', markersize=15)
    ax2.tick_params(axis='y', labelcolor='black')
    
    ax1.set_xlabel('X-Position(' + unit + ')')
    ax1.set_ylabel('Y-Position(' + unit + ')')
    ax2.set_ylabel('Y-Position(' + unit + ')')

    annots = []
    for ax in [ax1, ax2]:
        annot = ax1.annotate("", xy=(0, 0), xytext=(-20, 20), textcoords="offset points",
                             bbox=dict(boxstyle="round", fc="w", alpha=0.4),
                             arrowprops=dict(arrowstyle="->"))
        annot.set_visible(True)
        annots.append(annot)

    annot_dic = dict(zip([ax1, ax2], annots))
    line_dic = dict(zip([ax1, ax2], [line1, line2]))

    def update_annot(line, annot, ind):
        x, y = line.get_data()
        annot.xy = (x[ind["ind"][0]], y[ind["ind"][0]])
        text = "ID= {}".format(particle_ID_all[ind["ind"][0]])
        annot.set_text(text)

    def hover(event):

        if event.inaxes in [ax1, ax2]:
            for ax in [ax1, ax2]:
                cont, ind = line_dic[ax].contains(event)
                annot = annot_dic[ax]
                if cont:
                    update_annot(line_dic[ax], annot, ind)
                    annot.set_visible(True)
                    fig.canvas.draw_idle()
                else:
                    if annot.get_visible():
                        annot.set_visible(False)
                        fig.canvas.draw_idle()

    fig.canvas.mpl_connect("motion_notify_event", hover)
    ax1.set_box_aspect(1)
    ax2.set_box_aspect(1)

    plt.show()


def plot_bright_dark_psf_inTime(df_bright, df_dark, time_delay=0.1, dir_name=None):
    """
    Showing binding and unbinding events in time.

    Parameters
    ----------
    df_bright: pandas dataframe
        Bright PSF positions are stored in the data frame.

    df_dark: pandas dataframe
        Dark PSF positions are stored in the data frame.

    time_delay: float
        Define the time delay between binding and unbinding events frames. This only works when `flag_in_time` is set to True.

    dir_name: str
        You can save time slap frames if you specify a save path.
    """
    particle_ID_bright = df_bright['particle'].tolist()
    particle_x_bright = df_bright['x'].tolist()
    particle_y_bright = df_bright['y'].tolist()
    particle_size_bright = df_bright['bubble_size'].tolist()

    particle_ID_dark = df_dark['particle'].tolist()
    particle_x_dark = df_dark['x'].tolist()
    particle_y_dark = df_dark['y'].tolist()
    particle_size_dark = df_dark['bubble_size'].tolist()

    if len(particle_ID_bright) > len(particle_ID_dark):
        diff_len = len(particle_ID_bright) - len(particle_ID_dark)
        for _ in range(diff_len):
            particle_ID_dark.append(None)
            particle_x_dark.append(None)
            particle_y_dark.append(None)
            particle_size_dark.append(None)
    elif len(particle_ID_bright) < len(particle_ID_dark):
        diff_len = len(particle_ID_dark) - len(particle_ID_bright)
        for _ in range(diff_len):
            particle_ID_bright.append(None)
            particle_x_bright.append(None)
            particle_y_bright.append(None)
            particle_size_bright.append(None)

    fig, ax = plt.subplots(figsize=(5, 5))
    for i_ in range(len(particle_size_bright)):

        ax.scatter(particle_x_bright[i_], particle_y_bright[i_], label='1', color='blue', s=100, alpha=0.5, marker='o')
        ax.scatter(particle_x_dark[i_], particle_y_dark[i_], label='2', color='red', s=100, alpha=0.5, marker='o')

        ax.set_xlim(-0.1, 5.1)
        ax.set_ylim(-0.1, 5.1)

        if dir_name is not None:
            s_path = os.path.join(dir_name, 'bright_dark_psfs_frame' + str(i_) + '.png')
            plt.savefig(s_path)
        else:
            plt.pause(time_delay)
    plt.show()


def plot_histogram(data, title, fc='C1', ec='k'):

    plt.figure()
    plt.hist(data, alpha=0.7, bins=None, density=False, stacked=True, fc=fc, ec=ec)
    plt.title(title + '\nMedian: ' + str(np.median(data)))
    plt.ylabel('#Counts', fontsize=18)
    plt.xlabel('Length of linking', fontsize=18)
    plt.tight_layout()
    plt.show()




from __future__ import print_function
import numpy as np
from tqdm.autonotebook import tqdm


class ImagePatching():

    def __init__(self, depth=0, width=0, height=0, depth_overlap=0, width_overlap=0, height_overlap=0):

        self.total_start_position = None
        self.weight_matrix = None
        self.patches = None

        self.size = [depth, width, height, depth_overlap, width_overlap, height_overlap]
        self.depth_step = depth - depth_overlap
        self.width_step = width - width_overlap
        self.height_step = height - height_overlap

    def split_video(self, video):
        return [video[f:f+self.depth_step, w:w+self.width_step, h:h+self.height_step]
                for f in range(0, video.shape[0], self.depth_step)
                for w in range(0, video.shape[1], self.width_step)
                for h in range(0, video.shape[2], self.height_step)]

    def inverse_split(self, split_video, video_size):
        number_d_patch = int(video_size[0]/self.size[0])
        number_w_patch = int(video_size[1]/self.size[1])
        number_h_patch = int(video_size[2]/self.size[2])
        blocks = None
        index = 0
        for d_ in range(number_d_patch):
            temp1 = None

            for h_ in range(number_h_patch):
                temp0 = None

                for w_ in range(number_w_patch):

                    if temp0 is None:
                        temp0 = split_video[index]
                    else:
                        temp0 = np.concatenate((temp0, split_video[index]), axis=2)

                    if index <= len(split_video):
                        index = index + 1
                    else:
                        print('finsish!!!')
                        break
                if temp1 is None:
                    temp1 = temp0
                else:
                    temp1 = np.concatenate((temp1, temp0), axis=1)

            if blocks is None:
                blocks = temp1
            else:
                blocks = np.concatenate((blocks, temp1), axis=0)
        return blocks

    def split_weight_matrix(self, video, patch_size=16, strides=4):
        self.total_start_position = []
        self.weight_matrix = np.zeros((video.shape[1], video.shape[2]), dtype=np.int)
        self.patches = []
        for i_ in range(0, video.shape[1] - patch_size + strides, strides):
            for j_ in range(0, video.shape[2] - patch_size + strides, strides):
                sRow = i_
                eRow = i_ + patch_size

                sColumn = j_
                eColumn = j_ + patch_size

                list_start_position = [sRow, eRow, sColumn, eColumn]
                self.total_start_position.append(list_start_position)
                self.weight_matrix[sRow:eRow, sColumn:eColumn] = self.weight_matrix[sRow:eRow, sColumn:eColumn] + 1
                patch = video[:, sRow:eRow, sColumn:eColumn]
                self.patches.append(patch)
        return self.patches

    def reconstruction_weight_matrix(self, video_shape, new_patch):
        self.reconstruction_video = np.zeros(video_shape, dtype=np.float64)
        for i_ in tqdm(range(len(self.total_start_position))):
            sRow = self.total_start_position[i_][0]
            eRow = self.total_start_position[i_][1]

            sColumn = self.total_start_position[i_][2]
            eColumn = self.total_start_position[i_][3]

            self.reconstruction_video[:, sRow:eRow, sColumn:eColumn] = self.reconstruction_video[:, sRow:eRow, sColumn:eColumn] + new_patch[i_]

        self.weight_matrix_3D = np.broadcast_to(self.weight_matrix, (self.reconstruction_video.shape[0], video_shape[1], video_shape[2]))
        self.reconstruction_video = np.divide(self.reconstruction_video, self.weight_matrix_3D)
        return self.reconstruction_video

    def reconstruction_weight_matrix_kernel(self, i_):
        sRow = self.total_start_position[i_][0]
        eRow = self.total_start_position[i_][1]

        sColumn = self.total_start_position[i_][2]
        eColumn = self.total_start_position[i_][3]

        self.reconstruction_video[:, sRow:eRow, sColumn:eColumn] = self.reconstruction_video[:, sRow:eRow, sColumn:eColumn] + self.patches[i_]




